"""Testing utilities for components."""

from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional

from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.core.defs_module import load_yaml_component_from_path


def component_defs(
    *,
    component: Component,
    resources: Optional[Mapping[str, Any]] = None,
    context: Optional[ComponentLoadContext] = None,
) -> Definitions:
    """Builds a Definitions object from a Component.


    Args:
        component (Component): The Component to build the Definitions from.
        resources (Optional[Mapping[str, Any]]): A dictionary of resources to pass to the Component.
        context (Optional[ComponentLoadContext]): A ComponentLoadContext to pass to the Component. Defaults to a test context.

    Examples:

        .. code-block:: python

            # SimpleComponent produces an asset "an_asset"
            an_asset = component_defs(component=SimpleComponent()).get_assets_def("an_asset")
            assert an_asset.key == AssetKey("an_asset")

        .. code-block:: python

            # RequiresResoureComponent produces an asset "an_asset" and requires a resource "my_resource"
            an_asset = component_defs(
                component=SimpleComponent(),
                resources={"my_resource": MyResource()},
            ).get_assets_def("an_asset")
            assert an_asset.key == AssetKey("an_asset")
    """
    context = context or ComponentLoadContext.for_test()
    return component.build_defs(context).with_resources(resources)


def defs_from_component_yaml_path(
    *,
    component_yaml: Path,
    context: Optional[ComponentLoadContext] = None,
    resources: Optional[dict[str, Any]] = None,
):
    context = context or ComponentLoadContext.for_test()
    component = load_yaml_component_from_path(context=context, component_def_path=component_yaml)
    return component_defs(component=component, resources=resources, context=context)



from dagster_shared import record
from dagster_shared.serdes import whitelist_for_serdes

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.freshness import FreshnessState
from dagster._core.loader import LoadingContext
from dagster._streamline.asset_health import AssetHealthStatus


@whitelist_for_serdes
@record.record
class AssetFreshnessHealthState:
    """Maintains the latest freshness state for the asset."""

    freshness_state: FreshnessState

    @property
    def health_status(self) -> AssetHealthStatus:
        if self.freshness_state == FreshnessState.PASS:
            return AssetHealthStatus.HEALTHY
        elif self.freshness_state == FreshnessState.WARN:
            return AssetHealthStatus.WARNING
        elif self.freshness_state == FreshnessState.FAIL:
            return AssetHealthStatus.DEGRADED
        elif self.freshness_state == FreshnessState.NOT_APPLICABLE:
            return AssetHealthStatus.NOT_APPLICABLE
        else:
            return AssetHealthStatus.UNKNOWN

    @classmethod
    def compute_for_asset(
        cls, asset_key: AssetKey, loading_context: LoadingContext
    ) -> "AssetFreshnessHealthState":
        """Gets the freshness state for the asset from the DB."""
        freshness_state_record = loading_context.instance.get_entity_freshness_state(asset_key)

        if freshness_state_record is None:
            # freshness policy has no evaluations yet
            return cls(
                freshness_state=FreshnessState.UNKNOWN,
            )
        return cls(
            freshness_state=freshness_state_record.freshness_state,
        )



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Annotated, Literal, Optional, Union

from dagster_shared.record import as_dict, record

from dagster._core.definitions.asset_check_factories.metadata_bounds_checks import (
    build_metadata_bounds_checks,
)
from dagster._core.definitions.asset_check_factories.schema_change_checks import (
    build_column_schema_change_checks,
)
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.core_models import ResolvedAssetKey
from dagster.components.resolved.model import Resolver


def _resolve_asset_check_severity(context: ResolutionContext, model) -> AssetCheckSeverity:
    severity_str = context.resolve_value(model, as_type=str)
    return AssetCheckSeverity(severity_str)


ResolvedAssetCheckSeverity = Annotated[
    AssetCheckSeverity,
    Resolver(
        _resolve_asset_check_severity,
        model_field_type=str,
        description="The severity of the asset check.",
    ),
]


@record
class ColumnSchemaChangeParams(Resolvable):
    type: Literal["column_schema_change"]
    severity: ResolvedAssetCheckSeverity = AssetCheckSeverity.WARN

    def build_checks(self, key: AssetKey) -> Sequence[AssetChecksDefinition]:
        return build_column_schema_change_checks(
            assets=[key], **{k: v for k, v in as_dict(self).items() if k != "type"}
        )


@record
class MetadataBoundsParams(Resolvable):
    type: Literal["metadata_bounds"]
    metadata_key: str
    severity: ResolvedAssetCheckSeverity = AssetCheckSeverity.WARN
    min_value: Optional[Union[int, float]] = None
    max_value: Optional[Union[int, float]] = None
    exclusive_min: bool = False
    exclusive_max: bool = False

    def build_checks(self, key: AssetKey) -> Sequence[AssetChecksDefinition]:
        return build_metadata_bounds_checks(
            assets=[key], **{k: v for k, v in as_dict(self).items() if k != "type"}
        )


def resolve_check_params(
    context: ResolutionContext, model
) -> Union[ColumnSchemaChangeParams, MetadataBoundsParams]:
    if model.type == "column_schema_change":
        return ColumnSchemaChangeParams.resolve_from_model(context, model)
    elif model.type == "metadata_bounds":
        return MetadataBoundsParams.resolve_from_model(context, model)
    else:
        raise ValueError(f"Invalid check type: {model.type}")


# Doesn't work for union right now
ResolvedMetadataCheckParams = Annotated[
    Union[ColumnSchemaChangeParams, MetadataBoundsParams],
    Resolver(
        resolve_check_params,
        model_field_type=Union[ColumnSchemaChangeParams.model(), MetadataBoundsParams.model()],
    ),
]


@dataclass
class MetadataCheck(Resolvable):
    key: ResolvedAssetKey
    params: ResolvedMetadataCheckParams

    def build_checks(self) -> Sequence[AssetChecksDefinition]:
        return self.params.build_checks(self.key)


@dataclass
class MetadataChecksComponent(Component, Resolvable):
    checks: Sequence[MetadataCheck]

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return Definitions(
            asset_checks=[
                check for metadata_check in self.checks for check in metadata_check.build_checks()
            ]
        )



from abc import ABC
from collections.abc import Mapping
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Optional

from dagster_shared.serdes.utils import SerializableTimeDelta

from dagster._core.definitions.asset_key import AssetKey
from dagster._record import IHaveNew, record
from dagster._serdes import deserialize_value, whitelist_for_serdes
from dagster._utils import check


@whitelist_for_serdes
class FreshnessState(str, Enum):
    PASS = "PASS"
    WARN = "WARN"
    FAIL = "FAIL"
    UNKNOWN = "UNKNOWN"
    NOT_APPLICABLE = "NOT_APPLICABLE"


@whitelist_for_serdes
@record
class FreshnessStateEvaluation:
    key: AssetKey
    freshness_state: FreshnessState


@whitelist_for_serdes
@record
class FreshnessStateChange:
    """Event that is emitted when the freshness state of an asset changes."""

    key: AssetKey
    new_state: FreshnessState
    previous_state: FreshnessState
    state_change_timestamp: float


INTERNAL_FRESHNESS_POLICY_METADATA_KEY = "dagster/internal_freshness_policy"


class InternalFreshnessPolicy(ABC):
    @classmethod
    def from_asset_spec_metadata(
        cls, metadata: Mapping[str, Any]
    ) -> Optional["InternalFreshnessPolicy"]:
        serialized_policy = metadata.get(INTERNAL_FRESHNESS_POLICY_METADATA_KEY)
        if serialized_policy is None:
            return None
        return deserialize_value(serialized_policy.value, cls)  # pyright: ignore

    @staticmethod
    def time_window(
        fail_window: timedelta, warn_window: Optional[timedelta] = None
    ) -> "TimeWindowFreshnessPolicy":
        return TimeWindowFreshnessPolicy.from_timedeltas(fail_window, warn_window)


@whitelist_for_serdes
@record
class TimeWindowFreshnessPolicy(InternalFreshnessPolicy, IHaveNew):
    fail_window: SerializableTimeDelta
    warn_window: Optional[SerializableTimeDelta] = None

    @classmethod
    def from_timedeltas(cls, fail_window: timedelta, warn_window: Optional[timedelta] = None):
        check.invariant(
            fail_window.total_seconds() >= 60,
            "Due to Dagster system constraints, fail_window cannot be less than 1 minute",
        )
        if warn_window:
            check.invariant(
                warn_window.total_seconds() >= 60,
                "Due to Dagster system constraints, warn_window cannot be less than 1 minute",
            )
            check.invariant(warn_window < fail_window, "warn_window must be less than fail_window")

        return cls(
            fail_window=SerializableTimeDelta.from_timedelta(fail_window),
            warn_window=SerializableTimeDelta.from_timedelta(warn_window) if warn_window else None,
        )


@whitelist_for_serdes
@record
class FreshnessStateRecordBody:
    """Store serialized metadata about the freshness state for an entity.

    Left blank for now, a few examples of what we might want to store here:
    - Source timestamp for external assets / freshness checks
    - Snapshot of the freshness policy at the time of record creation
    """

    metadata: Optional[dict[str, Any]]


@record
class FreshnessStateRecord:
    entity_key: AssetKey
    freshness_state: FreshnessState
    updated_at: datetime
    record_body: FreshnessStateRecordBody

    @staticmethod
    def from_db_row(db_row):
        return FreshnessStateRecord(
            entity_key=check.not_none(AssetKey.from_db_string(db_row[0])),
            freshness_state=FreshnessState(db_row[3]),
            record_body=deserialize_value(db_row[4], FreshnessStateRecordBody),
            updated_at=db_row[5],
        )



from datetime import datetime
from typing import Optional

from dagster._record import record


@record
class TemporalContext:
    """TemporalContext represents an effective time, used for business logic, and last_event_id
    which is used to identify that state of the event log at some point in time. Put another way,
    the value of a TemporalContext represents a point in time and a snapshot of the event log.

    Effective time: This is the effective time of the computation in terms of business logic,
    and it impacts the behavior of partitioning and partition mapping. For example,
    the "last" partition window of a given partitions definition, it is with
    respect to the effective time.

    Last event id: Our event log has a monotonically increasing event id. This is used to
    cursor the event log. This event_id is also propogated to derived tables to indicate
    when that record is valid.  This allows us to query the state of the event log
    at a given point in time.

    Note that insertion time of the last_event_id is not the same as the effective time.

    A last_event_id of None indicates that the reads will be volatile and will immediately
    reflect any subsequent writes.
    """

    effective_dt: datetime
    last_event_id: Optional[int]



import logging
import sys
import threading
import time
from contextlib import ExitStack
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.instance import InstanceRef
from dagster._core.remote_representation.grpc_server_registry import GrpcServerRegistry
from dagster._core.remote_representation.origin import ManagedGrpcPythonEnvCodeLocationOrigin
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._grpc.__generated__ import dagster_api_pb2
from dagster._grpc.__generated__.dagster_api_pb2_grpc import DagsterApiServicer
from dagster._grpc.client import DEFAULT_GRPC_TIMEOUT
from dagster._grpc.constants import GrpcServerCommand
from dagster._grpc.types import (
    CancelExecutionRequest,
    CancelExecutionResult,
    ExecuteExternalJobArgs,
    SensorExecutionArgs,
    ShutdownServerResult,
    StartRunResult,
)
from dagster._serdes import deserialize_value, serialize_value
from dagster._utils.error import serializable_error_info_from_exc_info

if TYPE_CHECKING:
    from dagster._grpc.client import DagsterGrpcClient

CLEANUP_TICK = 1


class DagsterProxyApiServicer(DagsterApiServicer):
    """Service that implements the dagster gRPC API by opening up a "dagster api grpc" subprocess, and proxying
    all gRPC calls to the server running in that subprocess. This allows us to reload code by
    restarting the subprocess without needing to restart the parent process.
    """

    def __init__(
        self,
        loadable_target_origin: LoadableTargetOrigin,
        fixed_server_id: Optional[str],
        container_image: Optional[str],
        container_context: Optional[dict],
        inject_env_vars_from_instance: bool,
        location_name: Optional[str],
        log_level: str,
        startup_timeout: int,
        server_termination_event: threading.Event,
        instance_ref: Optional[InstanceRef],
        logger: logging.Logger,
        server_heartbeat: bool,
        server_heartbeat_timeout: int,
    ):
        super().__init__()

        self._loadable_target_origin = loadable_target_origin
        self._fixed_server_id = fixed_server_id
        self._container_image = container_image
        self._container_context = container_context
        self._inject_env_vars_from_instance = inject_env_vars_from_instance
        self._instance_ref = instance_ref
        self._location_name = location_name
        self._log_level = log_level
        self._logger = logger

        self._client = None
        self._load_error = None
        self._client_heartbeat_shutdown_event = None
        self._client_heartbeat_thread = None

        self._exit_stack = ExitStack()

        self._reload_lock = threading.Lock()

        self._grpc_server_registry = self._exit_stack.enter_context(
            GrpcServerRegistry(
                instance_ref=self._instance_ref,
                server_command=GrpcServerCommand.API_GRPC,
                heartbeat_ttl=30,
                startup_timeout=startup_timeout,
                log_level=self._log_level,
                inject_env_vars_from_instance=self._inject_env_vars_from_instance,
                container_image=self._container_image,
                container_context=self._container_context,
                wait_for_processes_on_shutdown=True,
                additional_timeout_msg="Set from --startup-timeout command line argument. ",
            )
        )
        self._origin = ManagedGrpcPythonEnvCodeLocationOrigin(
            loadable_target_origin=self._loadable_target_origin, location_name=location_name
        )

        self._server_termination_event = server_termination_event

        # Client tells the server to shutdown by calling ShutdownServer (or by failing to send a
        # hearbeat, at which point this event is set. The cleanup thread will then set the server
        # termination event once all current executions have finished, which will stop the server)
        self._shutdown_once_executions_finish_event = threading.Event()
        self.__cleanup_thread = threading.Thread(
            target=self._cleanup_thread,
            args=(),
            name="code-server-cleanup",
            daemon=True,
        )

        self.__last_heartbeat_time = time.time()
        self.__server_heartbeat_thread = None

        # Map runs to the client that launched them, so that we can route
        # termination requests
        self._run_clients: dict[str, DagsterGrpcClient] = {}

        self._reload_location()

        # Wait for the code server to have started before starting the heartbeat clock,
        # since the code loading and the server being ready is what will trigger the
        # heartbeats coming in from the client

        if server_heartbeat:
            self.__server_heartbeat_thread = threading.Thread(
                target=self._server_heartbeat_thread,
                args=(server_heartbeat_timeout,),
                name="grpc-server-heartbeat",
                daemon=True,
            )
            self.__server_heartbeat_thread.start()

        self.__cleanup_thread.start()

    def _reload_location(self):
        from dagster._grpc.client import client_heartbeat_thread

        try:
            endpoint = self._grpc_server_registry.reload_grpc_endpoint(self._origin)
            self._load_error = None
            self._client = endpoint.create_client()
        except Exception:
            # Server crashed when starting up
            self._load_error = serializable_error_info_from_exc_info(sys.exc_info())
            self._logger.exception("Failure while loading code")

        if self._client:
            self._client_heartbeat_shutdown_event = threading.Event()
            self._client_heartbeat_thread = threading.Thread(
                target=client_heartbeat_thread,
                args=(
                    self._client,
                    self._client_heartbeat_shutdown_event,
                ),
                name="grpc-client-heartbeat",
                daemon=True,
            )
            self._client_heartbeat_thread.start()

    def ReloadCode(self, request, context):
        with self._reload_lock:  # can only call this method once at a time
            old_heartbeat_shutdown_event = self._client_heartbeat_shutdown_event
            old_heartbeat_thread = self._client_heartbeat_thread
            old_client = self._client

            self._reload_location()  # Creates and starts a new heartbeat thread

        if old_client:
            old_client.shutdown_server()

        if old_heartbeat_shutdown_event:
            old_heartbeat_shutdown_event.set()

        if old_heartbeat_thread:
            old_heartbeat_thread.join()

        return dagster_api_pb2.ReloadCodeReply()

    def cleanup(self):
        # In case ShutdownServer was not called
        self._shutdown_once_executions_finish_event.set()
        self._grpc_server_registry.shutdown_all_processes()

        if self._client_heartbeat_shutdown_event:
            self._client_heartbeat_shutdown_event.set()
            self._client_heartbeat_shutdown_event = None

        if self._client_heartbeat_thread:
            self._client_heartbeat_thread.join()
            self._client_heartbeat_thread = None

        if self.__server_heartbeat_thread:
            self.__server_heartbeat_thread.join()
            self.__server_heartbeat_thread = None

        self._exit_stack.close()

        self.__cleanup_thread.join()

    def _cleanup_thread(self) -> None:
        while True:
            self._server_termination_event.wait(CLEANUP_TICK)
            if self._server_termination_event.is_set():
                break

            if self._shutdown_once_executions_finish_event.is_set():
                if self._grpc_server_registry.are_all_servers_shut_down():
                    self._server_termination_event.set()
                    self._grpc_server_registry.shutdown_all_processes()

    def _get_grpc_client(self):
        return self._client

    def _query(self, api_name: str, request, _context, timeout: int = DEFAULT_GRPC_TIMEOUT):
        if not self._client:
            raise Exception("No available client to code server")
        return check.not_none(self._client)._get_response(api_name, request, timeout)  # noqa

    def _server_heartbeat_thread(self, heartbeat_timeout: int) -> None:
        while True:
            if self._server_termination_event.is_set():
                break

            self._shutdown_once_executions_finish_event.wait(heartbeat_timeout)
            if self._shutdown_once_executions_finish_event.is_set():
                break

            if self.__last_heartbeat_time < time.time() - heartbeat_timeout:
                self._logger.warning(
                    f"No heartbeat received in {heartbeat_timeout} seconds, shutting down"
                )
                self._shutdown_once_executions_finish_event.set()
                self._grpc_server_registry.shutdown_all_processes()

    def _streaming_query(
        self, api_name: str, request, _context, timeout: int = DEFAULT_GRPC_TIMEOUT
    ):
        if not self._client:
            raise Exception("No available client to code server")
        return check.not_none(self._client)._get_streaming_response(api_name, request, timeout)  # noqa

    def ExecutionPlanSnapshot(self, request, context):
        return self._query("ExecutionPlanSnapshot", request, context)

    def ListRepositories(self, request, context):
        if self._load_error:
            return dagster_api_pb2.ListRepositoriesReply(
                serialized_list_repositories_response_or_error=serialize_value(self._load_error)
            )
        return self._query("ListRepositories", request, context)

    def Ping(self, request, context):
        return self._query("Ping", request, context)

    def GetServerId(self, request, context) -> dagster_api_pb2.GetServerIdReply:
        return (
            dagster_api_pb2.GetServerIdReply(server_id=self._fixed_server_id)
            if self._fixed_server_id
            else self._query("GetServerId", request, context)
        )

    def GetCurrentImage(self, request, context):
        return self._query("GetCurrentImage", request, context)

    def StreamingExternalRepository(self, request, context):
        return self._streaming_query("StreamingExternalRepository", request, context)

    def Heartbeat(self, request, context):
        self.__last_heartbeat_time = time.time()
        echo = request.echo
        return dagster_api_pb2.PingReply(echo=echo)

    def StreamingPing(self, request, context):
        return self._streaming_query("StreamingPing", request, context)

    def ExternalPartitionNames(self, request, context):
        return self._query("ExternalPartitionNames", request, context)

    def ExternalNotebookData(self, request, context):
        return self._query("ExternalNotebookData", request, context)

    def ExternalPartitionConfig(self, request, context):
        return self._query("ExternalPartitionConfig", request, context)

    def ExternalPartitionTags(self, request, context):
        return self._query("ExternalPartitionTags", request, context)

    def ExternalPartitionSetExecutionParams(self, request, context):
        return self._streaming_query("ExternalPartitionSetExecutionParams", request, context)

    def ExternalPipelineSubsetSnapshot(self, request, context):
        return self._query("ExternalPipelineSubsetSnapshot", request, context)

    def ExternalRepository(self, request, context):
        return self._query("ExternalRepository", request, context)

    def ExternalJob(self, request, context):
        return self._query("ExternalJob", request, context)

    def ExternalScheduleExecution(self, request, context):
        return self._streaming_query("ExternalScheduleExecution", request, context)

    def SyncExternalScheduleExecution(self, request, context):
        return self._query("SyncExternalScheduleExecution", request, context)

    def ExternalSensorExecution(self, request, context):
        sensor_execution_args = deserialize_value(
            request.serialized_external_sensor_execution_args,
            SensorExecutionArgs,
        )
        return self._streaming_query(
            "ExternalSensorExecution",
            request,
            context,
            sensor_execution_args.timeout or DEFAULT_GRPC_TIMEOUT,
        )

    def SyncExternalSensorExecution(self, request, context):
        sensor_execution_args = deserialize_value(
            request.serialized_external_sensor_execution_args,
            SensorExecutionArgs,
        )
        return self._query(
            "SyncExternalSensorExecution",
            request,
            context,
            sensor_execution_args.timeout or DEFAULT_GRPC_TIMEOUT,
        )

    def ShutdownServer(self, request, context):
        try:
            self._shutdown_once_executions_finish_event.set()
            self._grpc_server_registry.shutdown_all_processes()
            return dagster_api_pb2.ShutdownServerReply(
                serialized_shutdown_server_result=serialize_value(
                    ShutdownServerResult(success=True, serializable_error_info=None)
                )
            )
        except:
            return dagster_api_pb2.ShutdownServerReply(
                serialized_shutdown_server_result=serialize_value(
                    ShutdownServerResult(
                        success=False,
                        serializable_error_info=serializable_error_info_from_exc_info(
                            sys.exc_info()
                        ),
                    )
                )
            )

    def CancelExecution(self, request, context):
        try:
            cancel_execution_request = deserialize_value(
                request.serialized_cancel_execution_request,
                CancelExecutionRequest,
            )

            run_id = cancel_execution_request.run_id

            client = self._run_clients.get(run_id)
            if not client:
                raise Exception(f"Could not find a server for run {run_id}")

            return client._get_response("CancelExecution", request)  # noqa
        except:
            serializable_error_info = serializable_error_info_from_exc_info(sys.exc_info())

            return dagster_api_pb2.CancelExecutionReply(
                serialized_cancel_execution_result=serialize_value(
                    CancelExecutionResult(
                        success=False,
                        message=None,
                        serializable_error_info=serializable_error_info,
                    )
                )
            )

    def CanCancelExecution(self, request, context):
        return self._query("CanCancelExecution", request, context)

    def StartRun(self, request, context):
        if self._shutdown_once_executions_finish_event.is_set():
            return dagster_api_pb2.StartRunReply(
                serialized_start_run_result=serialize_value(
                    StartRunResult(
                        success=False,
                        message="Tried to start a run on a server after telling it to shut down",
                        serializable_error_info=None,
                    )
                )
            )

        execute_external_job_args = deserialize_value(
            request.serialized_execute_run_args,
            ExecuteExternalJobArgs,
        )

        run_id = execute_external_job_args.run_id

        client = self._client

        self._run_clients[run_id] = client  # pyright: ignore[reportArgumentType]
        return client._get_response("StartRun", request)  # noqa  # pyright: ignore[reportOptionalMemberAccess]



"""Facilities for running arbitrary commands in child processes."""

import os
import queue
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterator
from multiprocessing import Queue
from multiprocessing.context import BaseContext as MultiprocessingBaseContext
from multiprocessing.process import BaseProcess
from typing import TYPE_CHECKING, NamedTuple, Optional, Union

from typing_extensions import Literal

import dagster._check as check
from dagster._core.errors import DagsterExecutionInterruptedError
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster._utils.interrupts import capture_interrupts

if TYPE_CHECKING:
    from dagster._core.events import DagsterEvent


class ChildProcessEvent:
    pass


class ChildProcessStartEvent(
    NamedTuple("ChildProcessStartEvent", [("pid", int)]), ChildProcessEvent
):
    pass


class ChildProcessDoneEvent(NamedTuple("ChildProcessDoneEvent", [("pid", int)]), ChildProcessEvent):
    pass


class ChildProcessSystemErrorEvent(
    NamedTuple(
        "ChildProcessSystemErrorEvent", [("pid", int), ("error_info", SerializableErrorInfo)]
    ),
    ChildProcessEvent,
):
    pass


class ChildProcessCommand(ABC):
    """Inherit from this class in order to use this library.

    The object must be picklable; instantiate it and pass it to _execute_command_in_child_process.
    """

    @abstractmethod
    def execute(self) -> Iterator[Union[ChildProcessEvent, "DagsterEvent"]]:
        """This method is invoked in the child process.

        Yields a sequence of events to be handled by _execute_command_in_child_process.
        """


class ChildProcessCrashException(Exception):
    """Thrown when the child process crashes."""

    def __init__(self, pid, exit_code=None):
        self.pid = pid
        self.exit_code = exit_code
        super().__init__()


def _execute_command_in_child_process(event_queue: Queue, command: ChildProcessCommand):
    """Wraps the execution of a ChildProcessCommand.

    Handles errors and communicates across a queue with the parent process.
    """
    check.inst_param(command, "command", ChildProcessCommand)

    with capture_interrupts():
        pid = os.getpid()
        event_queue.put(ChildProcessStartEvent(pid=pid))
        try:
            for step_event in command.execute():
                event_queue.put(step_event)
            event_queue.put(ChildProcessDoneEvent(pid=pid))

        except (
            Exception,
            KeyboardInterrupt,
            DagsterExecutionInterruptedError,
        ):
            event_queue.put(
                ChildProcessSystemErrorEvent(
                    pid=pid, error_info=serializable_error_info_from_exc_info(sys.exc_info())
                )
            )


TICK = 20.0 * 1.0 / 1000.0
"""The minimum interval at which to check for child process liveness -- default 20ms."""

PROCESS_DEAD_AND_QUEUE_EMPTY = "PROCESS_DEAD_AND_QUEUE_EMPTY"
"""Sentinel value."""


def _poll_for_event(
    process, event_queue
) -> Optional[Union["DagsterEvent", Literal["PROCESS_DEAD_AND_QUEUE_EMPTY"]]]:
    try:
        return event_queue.get(block=True, timeout=TICK)
    except queue.Empty:
        if not process.is_alive():
            # There is a possibility that after the last queue.get the
            # process created another event and then died. In that case
            # we want to continue draining the queue.
            try:
                return event_queue.get(block=False)
            except queue.Empty:
                # If the queue empty we know that there are no more events
                # and that the process has died.
                return PROCESS_DEAD_AND_QUEUE_EMPTY
    return None


def execute_child_process_command(
    multiprocessing_ctx: MultiprocessingBaseContext, command: ChildProcessCommand
) -> Iterator[Optional[Union["DagsterEvent", ChildProcessEvent, BaseProcess]]]:
    """Execute a ChildProcessCommand in a new process.

    This function starts a new process whose execution target is a ChildProcessCommand wrapped by
    _execute_command_in_child_process; polls the queue for events yielded by the child process
    until the process dies and the queue is empty.

    This function yields a complex set of objects to enable having multiple child process
    executions in flight:
        * None - nothing has happened, yielded to enable cooperative multitasking other iterators

        * multiprocessing.BaseProcess - the child process object.

        * ChildProcessEvent - Family of objects that communicates state changes in the child process

        * The actual values yielded by the child process command

    Args:
        multiprocessing_ctx: The multiprocessing context to execute in (spawn, forkserver, fork)
        command (ChildProcessCommand): The command to execute in the child process.

    Warning: if the child process is in an infinite loop, this will
    also infinitely loop.
    """
    check.inst_param(command, "command", ChildProcessCommand)

    event_queue = multiprocessing_ctx.Queue()
    try:
        process = multiprocessing_ctx.Process(  # type: ignore
            target=_execute_command_in_child_process, args=(event_queue, command)
        )
        process.start()
        yield process

        completed_properly = False

        while not completed_properly:
            event = _poll_for_event(process, event_queue)

            if event == PROCESS_DEAD_AND_QUEUE_EMPTY:
                break

            yield event

            if isinstance(event, (ChildProcessDoneEvent, ChildProcessSystemErrorEvent)):
                completed_properly = True

        if not completed_properly:
            # TODO Figure out what to do about stderr/stdout
            raise ChildProcessCrashException(pid=process.pid, exit_code=process.exitcode)

        process.join()
    finally:
        event_queue.close()



import textwrap
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional, cast

import click
import yaml
from dagster_shared import check
from pydantic import BaseModel, TypeAdapter

from dagster.components.scaffold.scaffold import (
    NoParams,
    ScaffolderUnavailableReason,
    ScaffoldFormatOptions,
    ScaffoldRequest,
    get_scaffolder,
)


class ComponentDumper(yaml.Dumper):
    def write_line_break(self) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        # add an extra line break between top-level keys
        if self.indent == 0:
            super().write_line_break()
        super().write_line_break()

    # Format list elements with indentation - see https://github.com/yaml/pyyaml/issues/234
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)


def scaffold_component(
    request: ScaffoldRequest[Any],
    yaml_attributes: Optional[Mapping[str, Any]] = None,
) -> None:
    if request.scaffold_format == "yaml":
        with open(request.target_path / "defs.yaml", "w") as f:
            component_data = {"type": request.type_name, "attributes": yaml_attributes or {}}
            yaml.dump(
                component_data, f, Dumper=ComponentDumper, sort_keys=False, default_flow_style=False
            )
            f.writelines([""])
    elif request.scaffold_format == "python":
        with open(request.target_path / "component.py", "w") as f:
            fqtn = request.type_name
            check.invariant("." in fqtn, "Component must be a fully qualified type name")
            module_path, class_name = (
                ".".join(fqtn.split(".")[:-1]),
                fqtn.split(".")[-1],
            )
            f.write(
                textwrap.dedent(
                    f"""
                        from dagster import component, ComponentLoadContext
                        from {module_path} import {class_name}

                        @component_instance
                        def load(context: ComponentLoadContext) -> {class_name}: ...
                """
                ).lstrip()
            )
    else:
        check.assert_never(request.scaffold_format)


def scaffold_object(
    path: Path,
    obj: object,
    typename: str,
    json_params: Optional[str],
    scaffold_format: str,
    project_root: Optional[Path],
) -> None:
    from dagster.components.component.component import Component

    click.echo(f"Creating a component at {path}.")
    if not path.exists():
        path.mkdir(parents=True)

    scaffolder = get_scaffolder(obj)

    if isinstance(scaffolder, ScaffolderUnavailableReason):
        raise Exception(
            f"Object type {typename} does not have a scaffolder. Reason: {scaffolder.message}."
        )

    check.invariant(
        scaffold_format in ["yaml", "python"],
        f"scaffold must be either 'yaml' or 'python'. Got {scaffold_format}.",
    )

    params_model = parse_params_model(obj=obj, json_params=json_params)

    scaffolder.scaffold(
        ScaffoldRequest(
            type_name=typename,
            target_path=path,
            scaffold_format=cast("ScaffoldFormatOptions", scaffold_format),
            project_root=project_root,
            params=params_model,
        ),
    )

    if isinstance(obj, type) and issubclass(obj, Component):
        defs_yaml_path = path / "defs.yaml"
        component_py_path = path / "component.py"
        if not (defs_yaml_path.exists() or component_py_path.exists()):
            raise Exception(
                f"Currently all components require a defs.yaml or component.py file. Please ensure your implementation of scaffold writes this file at {defs_yaml_path} or {component_py_path}."
            )


def parse_params_model(obj: object, json_params: Optional[str]) -> BaseModel:
    scaffolder = get_scaffolder(obj)

    if isinstance(scaffolder, ScaffolderUnavailableReason):
        raise Exception(f"Object {obj} does not have a scaffolder. Reason: {scaffolder.message}.")

    # Get the params model class from the scaffolder
    params_model_cls = scaffolder.get_scaffold_params()

    if params_model_cls is NoParams and not (not json_params or json_params.strip() == "{}"):
        check.invariant(not json_params or json_params.strip() == "{}", "Input should be null")
        return NoParams()

    scaffold_params = TypeAdapter(params_model_cls).validate_json(json_params or "{}")
    json_params_dict = scaffold_params.model_dump()

    return params_model_cls.model_validate(json_params_dict or {})



import importlib
from functools import cached_property
from pathlib import Path
from types import ModuleType

from dagster_shared.record import record
from typing_extensions import TypeVar

from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext, use_component_load_context
from dagster.components.core.defs_module import DefsFolderComponent
from dagster.components.core.load_defs import get_library_json_enriched_defs

PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY = "plugin_component_types_json"

TComponent = TypeVar("TComponent", bound=Component)


@record(
    checked=False,  # cant handle ModuleType
)
class ComponentTree:
    defs_module: ModuleType
    project_root: Path

    @staticmethod
    def load(path_within_project: Path):
        """Using the provided path, find the nearest parent python project and load the
        ComponentTree using its configuration.
        """
        from dagster_dg_core.context import DgContext

        # replace with dagster_shared impl of path crawl and config resolution
        dg_context = DgContext.for_project_environment(path_within_project, command_line_config={})

        return ComponentTree(
            defs_module=importlib.import_module(dg_context.defs_module_name),
            project_root=dg_context.root_path,
        )

    @cached_property
    def load_context(self):
        return ComponentLoadContext.for_module(
            defs_module=self.defs_module,
            project_root=self.project_root,
        )

    @cached_property
    def root(self) -> DefsFolderComponent:
        return DefsFolderComponent.get(self.load_context)

    def load_defs(self) -> Definitions:
        with use_component_load_context(self.load_context):
            return Definitions.merge(
                self.root.build_defs(self.load_context),
                get_library_json_enriched_defs(),
            )

    def load_defs_at_path(self, defs_path: Path) -> Definitions:
        # impl to be fleshed out to flexibly handle different path types (str, list[str], ...)
        for cp, c in self.root.iterate_path_component_pairs():
            if cp.file_path.relative_to(self.root.path) == defs_path:
                ctx = self.load_context.for_path(cp.file_path)
                with use_component_load_context(ctx):
                    return c.build_defs(ctx)

        raise Exception(f"No component found for path {defs_path}")

    def get_all_components(
        self,
        of_type: type[TComponent],
    ) -> list[TComponent]:
        """Get all components from this context that are instance of the specified type."""
        return [
            component
            for component in self.root.iterate_components()
            if isinstance(component, of_type)
        ]



import os
import sys
import traceback
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Optional, TypeVar, Union, overload

from dagster_shared.yaml_utils.source_position import SourcePositionTree
from pydantic import BaseModel

from dagster._annotations import preview, public
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._record import copy, record
from dagster.components.resolved.errors import ResolutionException

T = TypeVar("T")


def env_scope(key: str) -> Optional[str]:
    return os.environ.get(key)


def automation_condition_scope() -> Mapping[str, Any]:
    return {
        "eager": AutomationCondition.eager,
        "on_cron": AutomationCondition.on_cron,
    }


T = TypeVar("T")


@public
@preview(emit_runtime_warning=False)
@record
class ResolutionContext:
    """The context available to Resolver functions when "resolving" from yaml in to a Resolvable object."""

    scope: Mapping[str, Any]
    path: list[Union[str, int]] = []
    source_position_tree: Optional[SourcePositionTree] = None

    def at_path(self, path_part: Union[str, int]):
        return copy(self, path=[*self.path, path_part])

    @staticmethod
    def default(source_position_tree: Optional[SourcePositionTree] = None) -> "ResolutionContext":
        return ResolutionContext(
            scope={"env": env_scope, "automation_condition": automation_condition_scope()},
            source_position_tree=source_position_tree,
        )

    def with_scope(self, **additional_scope) -> "ResolutionContext":
        return copy(self, scope={**self.scope, **additional_scope})

    def with_source_position_tree(
        self, source_position_tree: SourcePositionTree
    ) -> "ResolutionContext":
        return copy(self, source_position_tree=source_position_tree)

    def _location_parts(self, inline_err: str) -> list[str]:
        if not self.source_position_tree:
            return []

        source = self.source_position_tree.source_error(
            yaml_path=self.path,
            inline_error_message=inline_err,
            value_error=True,
        )
        return [
            f"{source.file_name}:{source.start_line_no} - {source.location}\n{source.snippet}\n",
        ]

    def _invalid_scope_exc(self, undefined_message: str) -> ResolutionException:
        msg_parts = self._location_parts(undefined_message)
        idx = undefined_message.find(" is undefined")
        if idx > 0:
            missing_scope = undefined_message[:idx]
            msg_parts.append(
                f"UndefinedError: {missing_scope} not found in scope, available scope is: {', '.join(self.scope.keys())}\n"
            )
        else:
            msg_parts.append(f"UndefinedError: {undefined_message}\n")

        return ResolutionException("".join(msg_parts))

    def _scope_threw_exc(self, fmt_exc: list[str]):
        msg_parts = self._location_parts(fmt_exc[-1] if fmt_exc else "ResolutionException")
        msg_parts.append("Exception occurred while evaluating template.\n")
        # strip the system frames from the stack trace
        template_seen = False
        for line in fmt_exc:
            if "in top-level template code" in line:
                template_seen = True
            if line.strip().startswith("File") and not template_seen:
                continue
            msg_parts.append(line)

        return ResolutionException("".join(msg_parts))

    def build_resolve_fn_exc(
        self,
        fmt_exc: list[str],
        field_name: str,
        model: BaseModel,
    ) -> ResolutionException:
        msg_parts = self._location_parts(fmt_exc[-1] if fmt_exc else "ResolutionException")
        msg_parts.append(
            f"Exception occurred in Resolver for field '{field_name}' resolving from {model.__class__.__name__}.\n"
        )
        msg_parts.extend(fmt_exc)

        return ResolutionException("".join(msg_parts))

    def _resolve_inner_value(self, val: Any) -> Any:
        """Resolves a single value, if it is a templated string."""
        # defer for import performance
        from jinja2 import Undefined
        from jinja2.exceptions import UndefinedError
        from jinja2.nativetypes import NativeTemplate

        if isinstance(val, str):
            try:
                val = NativeTemplate(val).render(**self.scope)
                if isinstance(val, Undefined):
                    raise self._invalid_scope_exc(val._undefined_message) from None  # noqa: SLF001
                return val
            except UndefinedError as undefined_error:
                if undefined_error.message:
                    raise self._invalid_scope_exc(undefined_error.message) from None
                else:
                    raise
            except Exception as e:
                if not isinstance(e, ResolutionException):
                    fmt_exc = traceback.format_exception(*sys.exc_info())
                    raise self._scope_threw_exc(fmt_exc) from None
                else:
                    raise
        else:
            return val

    @overload
    def resolve_value(self, val: Any, as_type: type[T]) -> T: ...

    @overload
    def resolve_value(self, val: Mapping) -> Mapping: ...

    @overload
    def resolve_value(self, val: tuple) -> tuple: ...

    @overload
    def resolve_value(self, val: Sequence) -> Sequence: ...

    def resolve_value(self, val: Any, as_type: Optional[type] = None) -> Any:
        """Recursively resolves templated values in a nested object."""
        if isinstance(val, dict):
            return {k: self.at_path(k).resolve_value(v) for k, v in val.items()}
        elif isinstance(val, tuple):
            return tuple(self.at_path(i).resolve_value(v) for i, v in enumerate(val))
        elif isinstance(val, list):
            return [self.at_path(i).resolve_value(v) for i, v in enumerate(val)]
        else:
            return self._resolve_inner_value(val)

    def resolve_source_relative_path(self, value: str) -> Path:
        path = Path(value)
        if not self.source_position_tree or path.is_absolute():
            # no source, can't transform
            return path

        source_dir = Path(self.source_position_tree.position.filename).parent
        return source_dir.joinpath(path).resolve()



import textwrap
from typing import Optional

from pydantic import BaseModel

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster.components.lib.shim_components.base import ShimScaffolder
from dagster.components.scaffold.scaffold import ScaffoldRequest, scaffold_with


class MultiAssetScaffoldParams(BaseModel):
    asset_key: Optional[list[str]] = None


class MultiAssetScaffolder(ShimScaffolder[MultiAssetScaffoldParams]):
    @classmethod
    def get_scaffold_params(cls) -> type[MultiAssetScaffoldParams]:
        return MultiAssetScaffoldParams

    def get_text(self, request: ScaffoldRequest[MultiAssetScaffoldParams]) -> str:
        asset_keys = (
            request.params.asset_key
            if request.params and request.params.asset_key
            # Default to two sample assets based on the filename
            else [
                f"{request.target_path.stem}/first_asset",
                f"{request.target_path.stem}/second_asset",
            ]
        )

        specs_str = textwrap.indent(
            ",\n".join(
                f"dg.AssetSpec(key=dg.AssetKey({AssetKey.from_user_string(key).path!r}))"
                for key in asset_keys
            ),
            prefix=" " * 20,
        )
        return textwrap.dedent(
            f"""\
            import dagster as dg


            @dg.multi_asset(
                specs=[
{specs_str}
                ]
            )
            def {request.target_path.stem}(context: dg.AssetExecutionContext):
                ...
            """
        )


scaffold_with(MultiAssetScaffolder)(multi_asset)



import importlib
from pathlib import Path
from types import ModuleType
from typing import Optional

from dagster_shared.serdes.objects.package_entry import json_for_all_components

from dagster._annotations import deprecated, preview, public
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils.warnings import suppress_dagster_warnings
from dagster.components.core.context import ComponentLoadContext, use_component_load_context

PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY = "plugin_component_types_json"


@deprecated(breaking_version="0.2.0")
@suppress_dagster_warnings
def build_component_defs(components_root: Path) -> Definitions:
    """Build a Definitions object for all the component instances in a given code location.

    Args:
        components_root (Path): The path to the components root. This is a directory containing
            subdirectories with component instances.
    """
    defs_root = importlib.import_module(
        f"{Path(components_root).parent.name}.{Path(components_root).name}"
    )

    return load_defs(defs_root=defs_root, project_root=components_root.parent.parent)


def get_project_root(defs_root: ModuleType) -> Path:
    """Find the project root directory containing pyproject.toml or setup.py.

    Args:
        defs_root: A module object from which to start the search.

    Returns:
        The absolute path to the project root directory.

    Raises:
        FileNotFoundError: If no project root with pyproject.toml or setup.py is found.
    """
    # Get the module's file path

    module_path = getattr(defs_root, "__file__", None)
    if not module_path:
        raise FileNotFoundError(f"Module {defs_root} has no __file__ attribute")

    # Start with the directory containing the module
    current_dir = Path(module_path).parent

    # Traverse up until we find pyproject.toml or setup.py
    while current_dir != current_dir.parent:  # Stop at root
        if (current_dir / "pyproject.toml").exists() or (current_dir / "setup.py").exists():
            return current_dir
        current_dir = current_dir.parent

    raise FileNotFoundError("No project root with pyproject.toml or setup.py found")


# Public method so optional Nones are fine
@public
@preview(emit_runtime_warning=False)
@suppress_dagster_warnings
def load_defs(defs_root: ModuleType, project_root: Optional[Path] = None) -> Definitions:
    """Constructs a Definitions object, loading all Dagster defs in the given module.

    Args:
        defs_root (Path): The path to the defs root, typically `package.defs`.
        project_root (Optional[Path]): path to the project root directory.
    """
    from dagster.components.core.defs_module import get_component

    project_root = project_root if project_root else get_project_root(defs_root)

    # create a top-level DefsModule component from the root module
    context = ComponentLoadContext.for_module(defs_root, project_root)
    with use_component_load_context(context):
        root_component = get_component(context)
        if root_component is None:
            raise DagsterInvalidDefinitionError("Could not resolve root module to a component.")

        return Definitions.merge(
            root_component.build_defs(context),
            get_library_json_enriched_defs(),
        )


def get_library_json_enriched_defs() -> Definitions:
    from dagster.components.core.package_entry import discover_entry_point_package_objects
    from dagster.components.core.snapshot import get_package_entry_snap

    library_objects = discover_entry_point_package_objects()
    snaps = [get_package_entry_snap(key, obj) for key, obj in library_objects.items()]
    components_json = json_for_all_components(snaps)

    return Definitions(metadata={PLUGIN_COMPONENT_TYPES_JSON_METADATA_KEY: components_json})



from collections.abc import Mapping, Sequence
from typing import Any, Optional, Union

import dagster._check as check
from dagster._config import ConfigSchemaSnapshot
from dagster._core.snap import DependencyStructureIndex, JobSnap
from dagster._core.snap.dagster_types import DagsterTypeSnap
from dagster._core.snap.mode import ModeDefSnap
from dagster._core.snap.node import GraphDefSnap, OpDefSnap


class JobIndex:
    job_snapshot: JobSnap
    parent_job_snapshot: Optional[JobSnap]
    _node_defs_snaps_index: Mapping[str, Union[OpDefSnap, GraphDefSnap]]
    _dagster_type_snaps_by_name_index: Mapping[str, DagsterTypeSnap]
    dep_structure_index: DependencyStructureIndex
    _comp_dep_structures: Mapping[str, DependencyStructureIndex]
    _job_snapshot_id: Optional[str]

    def __init__(
        self,
        job_snapshot: JobSnap,
        parent_job_snapshot: Optional[JobSnap],
    ):
        self.job_snapshot = check.inst_param(job_snapshot, "job_snapshot", JobSnap)
        self.parent_job_snapshot = check.opt_inst_param(
            parent_job_snapshot, "parent_job_snapshot", JobSnap
        )

        node_def_snaps: Sequence[Union[OpDefSnap, GraphDefSnap]] = [
            *job_snapshot.node_defs_snapshot.op_def_snaps,
            *job_snapshot.node_defs_snapshot.graph_def_snaps,
        ]
        self._node_defs_snaps_index = {sd.name: sd for sd in node_def_snaps}

        self._dagster_type_snaps_by_name_index = {
            dagster_type_snap.name: dagster_type_snap
            for dagster_type_snap in job_snapshot.dagster_type_namespace_snapshot.all_dagster_type_snaps_by_key.values()
            if dagster_type_snap.name
        }

        self.dep_structure_index = DependencyStructureIndex(job_snapshot.dep_structure_snapshot)

        self._comp_dep_structures = {
            comp_snap.name: DependencyStructureIndex(comp_snap.dep_structure_snapshot)
            for comp_snap in job_snapshot.node_defs_snapshot.graph_def_snaps
        }

    @property
    def name(self) -> str:
        return self.job_snapshot.name

    @property
    def description(self) -> Optional[str]:
        return self.job_snapshot.description

    @property
    def tags(self) -> Mapping[str, Any]:
        return self.job_snapshot.tags

    @property
    def metadata(self):
        return self.job_snapshot.metadata

    @property
    def job_snapshot_id(self) -> str:
        return self.job_snapshot.snapshot_id

    def has_dagster_type_name(self, type_name: str) -> bool:
        return type_name in self._dagster_type_snaps_by_name_index

    def get_dagster_type_from_name(self, type_name: str) -> DagsterTypeSnap:
        return self._dagster_type_snaps_by_name_index[type_name]

    def get_node_def_snap(self, node_def_name: str) -> Union[OpDefSnap, GraphDefSnap]:
        check.str_param(node_def_name, "node_def_name")
        return self._node_defs_snaps_index[node_def_name]

    def get_dep_structure_index(self, graph_def_name: str) -> DependencyStructureIndex:
        return self._comp_dep_structures[graph_def_name]

    def get_dagster_type_snaps(self) -> Sequence[DagsterTypeSnap]:
        dt_namespace = self.job_snapshot.dagster_type_namespace_snapshot
        return list(dt_namespace.all_dagster_type_snaps_by_key.values())

    def has_node_invocation(self, node_name: str) -> bool:
        return self.dep_structure_index.has_invocation(node_name)

    def get_default_mode_name(self) -> str:
        return self.job_snapshot.mode_def_snaps[0].name

    def has_mode_def(self, name: str) -> bool:
        check.str_param(name, "name")
        for mode_def_snap in self.job_snapshot.mode_def_snaps:
            if mode_def_snap.name == name:
                return True

        return False

    @property
    def available_modes(self) -> Sequence[str]:
        return [mode_def_snap.name for mode_def_snap in self.job_snapshot.mode_def_snaps]

    def get_mode_def_snap(self, name: str) -> ModeDefSnap:
        check.str_param(name, "name")
        for mode_def_snap in self.job_snapshot.mode_def_snaps:
            if mode_def_snap.name == name:
                return mode_def_snap

        check.failed(f"Mode {name} not found")

    @property
    def config_schema_snapshot(self) -> ConfigSchemaSnapshot:
        return self.job_snapshot.config_schema_snapshot



import enum

from dagster_shared.serdes import whitelist_for_serdes


@whitelist_for_serdes
class AssetHealthStatus(enum.Enum):
    """Enum for the health status of an asset."""

    HEALTHY = "HEALTHY"
    WARNING = "WARNING"
    DEGRADED = "DEGRADED"
    UNKNOWN = "UNKNOWN"
    NOT_APPLICABLE = "NOT_APPLICABLE"



import importlib.util
import subprocess
import sys
import textwrap
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, TypeVar, Union

import click
from pydantic import BaseModel

from dagster import _check as check
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.errors import DagsterError
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.core_models import (
    AssetAttributesModel,
    resolve_asset_spec_update_kwargs_to_mapping,
)

T = TypeVar("T")


def exit_with_error(error_message: str) -> None:
    click.echo(click.style(error_message, fg="red"))
    sys.exit(1)


def format_error_message(message: str) -> str:
    # width=10000 unwraps any hardwrapping
    dedented = textwrap.dedent(message).strip()
    paragraphs = [textwrap.fill(p, width=10000) for p in dedented.split("\n\n")]
    return "\n\n".join(paragraphs)


# Temporarily places a path at the front of sys.path, ensuring that any modules in that path are
# importable.
@contextmanager
def ensure_loadable_path(path: Path) -> Iterator[None]:
    orig_path = sys.path.copy()
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path = orig_path


def get_path_for_package(package_name: str) -> str:
    spec = importlib.util.find_spec(package_name)
    if not spec:
        raise DagsterError(f"Cannot find package: {package_name}")
    # file_path = spec.origin
    submodule_search_locations = spec.submodule_search_locations
    if not submodule_search_locations:
        raise DagsterError(f"Package does not have any locations for submodules: {package_name}")
    return submodule_search_locations[0]


TRANSLATOR_MERGE_ATTRIBUTES = {"metadata", "tags"}


@dataclass
class TranslatorResolvingInfo:
    obj_name: str
    asset_attributes: Union[str, BaseModel]
    resolution_context: ResolutionContext
    model_key: str = "asset_attributes"

    def _resolve_asset_attributes(self, context: Mapping[str, Any]) -> Union[AssetSpec, BaseModel]:
        """Resolves the user-specified asset attributes into an AssetAttributesModel, or an AssetSpec
        if the UDF returns one.
        """
        if not isinstance(self.asset_attributes, str):
            return self.asset_attributes

        resolved_asset_attributes = (
            self.resolution_context.at_path(self.model_key)
            .with_scope(**context)
            .resolve_value(self.asset_attributes)
        )

        if isinstance(resolved_asset_attributes, AssetSpec):
            return resolved_asset_attributes
        elif isinstance(resolved_asset_attributes, AssetAttributesModel):
            return resolved_asset_attributes
        elif isinstance(resolved_asset_attributes, dict):
            return AssetAttributesModel(**(resolved_asset_attributes))
        else:
            check.failed(
                f"Unexpected return value for asset_attributes UDF: {type(resolved_asset_attributes)}"
            )

    def get_asset_spec(self, base_spec: AssetSpec, context: Mapping[str, Any]) -> AssetSpec:
        """Returns an AssetSpec that combines the base spec with attributes resolved using the provided context.

        Usage:

        ```python
        class WrappedDagsterXTranslator(DagsterXTranslator):
            def __init__(self, *, base_translator, resolving_info: TranslatorResolvingInfo):
                self.base_translator = base_translator
                self.resolving_info = resolving_info

            def get_asset_spec(self, base_spec: AssetSpec, x_params: Any) -> AssetSpec:
                return self.resolving_info.get_asset_spec(
                    base_spec, {"x_params": x_params}
                )

        ```
        """
        resolved_asset_attributes = self._resolve_asset_attributes(context)
        if isinstance(resolved_asset_attributes, AssetSpec):
            return resolved_asset_attributes

        resolved_attributes = dict(
            resolve_asset_spec_update_kwargs_to_mapping(
                model=resolved_asset_attributes,
                context=self.resolution_context.at_path(self.model_key).with_scope(**context),
            )
        )
        if "code_version" in resolved_attributes:
            resolved_attributes = {
                **resolved_attributes,
                "code_version": str(resolved_attributes["code_version"]),
            }

        if "key_prefix" in resolved_attributes:
            prefix = resolved_attributes.pop("key_prefix")
            if "key" in resolved_attributes:
                key = resolved_attributes.pop("key")
            else:
                key = base_spec.key
            key = key.with_prefix(prefix)
            resolved_attributes["key"] = key

        return base_spec.replace_attributes(
            **{k: v for k, v in resolved_attributes.items() if k not in TRANSLATOR_MERGE_ATTRIBUTES}
        ).merge_attributes(
            **{k: v for k, v in resolved_attributes.items() if k in TRANSLATOR_MERGE_ATTRIBUTES}
        )


# ########################
# ##### PLATFORM
# ########################


def is_windows() -> bool:
    return sys.platform == "win32"


def is_macos() -> bool:
    return sys.platform == "darwin"


# ########################
# ##### VENV
# ########################


def get_venv_executable(venv_dir: Path, executable: str = "python") -> Path:
    if is_windows():
        return venv_dir / "Scripts" / f"{executable}.exe"
    else:
        return venv_dir / "bin" / executable


def install_to_venv(venv_dir: Path, install_args: list[str]) -> None:
    executable = get_venv_executable(venv_dir)
    command = ["uv", "pip", "install", "--python", str(executable), *install_args]
    subprocess.run(command, check=True)


def get_path_from_module(module: ModuleType) -> Path:
    module_path = (
        Path(module.__file__).parent
        if module.__file__
        else Path(module.__path__[0])
        if module.__path__
        else None
    )
    return check.not_none(module_path, f"Module {module.__name__} has no filepath")



from typing import TYPE_CHECKING, Any, Callable, TypeVar

from dagster._annotations import preview, public, superseded

if TYPE_CHECKING:
    from dagster.components.component.component import Component
    from dagster.components.core.context import ComponentLoadContext

COMPONENT_LOADER_FN_ATTR = "__dagster_component_loader_fn"


T_Component = TypeVar("T_Component", bound="Component")


@public
@superseded(additional_warn_text="Use `component_instance` instead.")
def component(
    fn: Callable[["ComponentLoadContext"], T_Component],
) -> Callable[["ComponentLoadContext"], T_Component]:
    """Decorator for a function to be used to load an instance of a Component.
    This is used when instantiating components in python instead of via yaml.
    """
    return component_instance(fn)


@public
@preview(emit_runtime_warning=False)
def component_instance(
    fn: Callable[["ComponentLoadContext"], T_Component],
) -> Callable[["ComponentLoadContext"], T_Component]:
    """Decorator for a function to be used to load an instance of a Component.
    This is used when instantiating components in python instead of via yaml.
    """
    setattr(fn, COMPONENT_LOADER_FN_ATTR, True)
    return fn


def is_component_loader(obj: Any) -> bool:
    return getattr(obj, COMPONENT_LOADER_FN_ATTR, False)



from dagster_shared import record
from dagster_shared.serdes import whitelist_for_serdes

from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_key import AssetCheckKey
from dagster._core.loader import LoadingContext
from dagster._streamline.asset_health import AssetHealthStatus


@whitelist_for_serdes
@record.record
class AssetCheckHealthState:
    """Maintains a list of asset checks for the asset in each terminal state. If a check is in progress,
    it will not move to a new list until the execution is complete.
    """

    passing_checks: set[AssetCheckKey]
    failing_checks: set[AssetCheckKey]
    warning_checks: set[AssetCheckKey]
    all_checks: set[AssetCheckKey]

    @classmethod
    def default(cls) -> "AssetCheckHealthState":
        return cls(
            passing_checks=set(),
            failing_checks=set(),
            warning_checks=set(),
            all_checks=set(),
        )

    @property
    def health_status(self) -> AssetHealthStatus:
        """Returns the health status of the asset based on the checks."""
        if len(self.all_checks) == 0:
            return AssetHealthStatus.NOT_APPLICABLE
        if len(self.failing_checks) > 0:
            return AssetHealthStatus.DEGRADED
        if len(self.warning_checks) > 0:
            return AssetHealthStatus.WARNING

        num_unexecuted = (
            len(self.all_checks)
            - len(self.passing_checks)
            - len(self.failing_checks)
            - len(self.warning_checks)
        )
        if num_unexecuted > 0:
            return AssetHealthStatus.UNKNOWN
        # all checks are passing
        return AssetHealthStatus.HEALTHY

    @classmethod
    async def compute_for_asset_checks(
        cls, check_keys: set[AssetCheckKey], loading_context: LoadingContext
    ) -> "AssetCheckHealthState":
        """Using the latest terminal state for each check as stored in the DB, returns a set of
        asset checks in each terminal state. If a check is in progress, it remains in the terminal
        state it was in prior to the in progress execution.
        """
        from dagster._core.storage.asset_check_execution_record import (
            AssetCheckExecutionResolvedStatus,
        )
        from dagster._core.storage.event_log.base import AssetCheckSummaryRecord

        if len(check_keys) == 0:
            # no checks defined
            return AssetCheckHealthState.default()

        passing_checks = set()
        warning_checks = set()
        failing_checks = set()

        check_records = await AssetCheckSummaryRecord.gen_many(
            loading_context,
            check_keys,
        )

        for check_record in check_records:
            if check_record is None or check_record.last_check_execution_record is None:
                # check has never been executed
                continue

            check_key = check_record.asset_check_key

            # if the last_check_execution_record is completed, it will be the same as last_completed_check_execution_record,
            # but we check the last_check_execution_record status first since there is an edge case
            # where the record will have status PLANNED, but the resolve_status will be EXECUTION_FAILED
            # because the run for the check failed.
            last_check_execution_status = (
                await check_record.last_check_execution_record.resolve_status(loading_context)
            )
            last_check_evaluation = check_record.last_check_execution_record.evaluation

            if last_check_execution_status in [
                AssetCheckExecutionResolvedStatus.IN_PROGRESS,
                AssetCheckExecutionResolvedStatus.SKIPPED,
            ]:
                # the last check is still in progress or is skipped, so we want to check the status of
                # the latest completed check instead
                if check_record.last_completed_check_execution_record is None:
                    # the check hasn't been executed prior to this in progress check
                    continue
                # should never need to fetch a run since the non-resolved status is success or failed.
                # this method converts directly to the resolved status
                last_check_execution_status = (
                    await check_record.last_completed_check_execution_record.resolve_status(
                        loading_context
                    )
                )
                last_check_evaluation = (
                    check_record.last_completed_check_execution_record.evaluation
                )

            if last_check_execution_status == AssetCheckExecutionResolvedStatus.FAILED:
                # failed checks should always have an evaluation, but default to ERROR if not
                if (
                    last_check_evaluation
                    and last_check_evaluation.severity == AssetCheckSeverity.WARN
                ):
                    warning_checks.add(check_key)
                else:
                    failing_checks.add(check_key)
            elif last_check_execution_status == AssetCheckExecutionResolvedStatus.EXECUTION_FAILED:
                # EXECUTION_FAILED checks don't have an evaluation and we want to report them as failures
                failing_checks.add(check_key)
            else:
                # asset check passed
                passing_checks.add(check_key)

        return AssetCheckHealthState(
            passing_checks=passing_checks,
            failing_checks=failing_checks,
            warning_checks=warning_checks,
            all_checks=check_keys,
        )



import inspect
from typing import Callable, Generic, Optional, TypeVar, Union, cast

from dagster._annotations import preview, public
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.errors import DagsterInvariantViolationError
from dagster._record import record
from dagster.components.core.context import ComponentLoadContext

T_Defs = TypeVar("T_Defs", Definitions, RepositoryDefinition)


@record
class LazyDefinitions(Generic[T_Defs]):
    """An object that can be invoked to load a set of definitions. Useful in tests when you want to regenerate the same definitions in multiple contexts."""

    load_fn: Callable[..., T_Defs]
    has_context_arg: bool

    def __call__(self, context: Optional[ComponentLoadContext] = None) -> T_Defs:
        """Load a set of definitions using the load_fn provided at construction time.

        Args:
            context (Optional[ComponentLoadContext]): Optional context for loading definitions.

        Returns:
            T_Defs: The loaded definitions.
        """
        if self.has_context_arg:
            if context is None:
                raise DagsterInvariantViolationError(
                    "Function requires a ComponentLoadContext but none was provided"
                )
            result = self.load_fn(context)
        else:
            result = self.load_fn()

        if not isinstance(result, (Definitions, RepositoryDefinition)):
            raise DagsterInvariantViolationError(
                "Function must return a Definitions or RepositoryDefinition object"
            )
        return result


@public
@preview(emit_runtime_warning=False)
def definitions(
    fn: Union[Callable[[], Definitions], Callable[[ComponentLoadContext], Definitions]],
) -> Callable[..., Definitions]:
    """Marks a function as an entry point for loading a set of Dagster definitions. It is an alternative
    to directly instantiating a Definitions object and assigning it to a local variable. This enables
    a user to import a python module that contains a loadable definitions object without having
    to create it at import time.

    The function can optionally accept a ComponentLoadContext parameter. If it does, the context will be
    passed to the function when it is called. If it doesn't, the function will be called without any
    parameters.

    Returns:
        Callable[..., Definitions]: A callable that will load a set of definitions when invoked.
            The callable accepts an optional ComponentLoadContext parameter that defaults to None.

    Examples:
        .. code-block:: python

            import dagster as dg

            # Example with context parameter
            @dg.definitions
            def defs_with_context(context: ComponentLoadContext):
                @asset
                def regular_asset(): ...

                return Definitions(assets=[regular_asset])

            # Example without context parameter
            @dg.definitions
            def defs_without_context():
                @asset
                def regular_asset(): ...

                return Definitions(assets=[regular_asset])

    """
    sig = inspect.signature(fn)
    has_context_arg = False

    if len(sig.parameters) > 0:
        first_param = next(iter(sig.parameters.values()))
        if first_param.annotation == ComponentLoadContext:
            has_context_arg = True
            if len(sig.parameters) > 1:
                raise DagsterInvariantViolationError(
                    "Function must accept either no parameters or exactly one ComponentLoadContext parameter"
                )
        else:
            raise DagsterInvariantViolationError(
                "Function must accept either no parameters or exactly one ComponentLoadContext parameter"
            )

    lazy_defs = LazyDefinitions[Definitions](load_fn=fn, has_context_arg=has_context_arg)
    return cast("Callable[..., Definitions]", lazy_defs)


# For backwards compatibility with existing test cases
def lazy_repository(fn: Callable[[], RepositoryDefinition]) -> Callable[[], RepositoryDefinition]:
    lazy_defs = LazyDefinitions[RepositoryDefinition](load_fn=fn, has_context_arg=False)
    return cast("Callable[[],RepositoryDefinition]", lazy_defs)



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Iterator, Optional, Sequence

from dagster._core.errors import DagsterInvariantViolationError
from dagster._serdes import whitelist_for_serdes
from dagster_shared.record import LegacyNamedTupleMixin, record, as_dict


@whitelist_for_serdes
@record
class LoadableTargetOrigin(LegacyNamedTupleMixin):
    executable_path: Optional[str] = None
    python_file: Optional[str] = None
    module_name: Optional[str] = None
    working_directory: Optional[str] = None
    attribute: Optional[str] = None
    package_name: Optional[str] = None

    def get_cli_args(self) -> Sequence[str]:
        args = (
            (["-f", self.python_file] if self.python_file else [])
            + (["-m", self.module_name] if self.module_name else [])
            + (["-d", self.working_directory] if self.working_directory else [])
            + (["-a", self.attribute] if self.attribute else [])
            + (["--package-name", self.package_name] if self.package_name else [])
        )

        return args

    @staticmethod
    def get() -> "LoadableTargetOrigin":
        ctx = _current_loadable_target_origin.get(None)
        if ctx is None:
            raise DagsterInvariantViolationError(
                "No LoadableTargetOrigin is currently being loaded."
            )
        return ctx

    @property
    def as_dict(self) -> dict:
        return {k: v for k, v in as_dict(self).items() if v is not None}


_current_loadable_target_origin: ContextVar[Optional[LoadableTargetOrigin]] = (
    ContextVar("_current_loadable_target_origin", default=None)
)


@contextmanager
def enter_loadable_target_origin_load_context(
    loadable_target_origin: LoadableTargetOrigin,
) -> Iterator[None]:
    token = _current_loadable_target_origin.set(loadable_target_origin)
    try:
        yield
    finally:
        _current_loadable_target_origin.reset(token)



import re
from typing import NamedTuple, Optional, Union, cast

import dagster._check as check
from dagster._core.definitions.dependency import NodeHandle
from dagster._serdes import whitelist_for_serdes


# Serialize node_handle -> solid_handle for backcompat
@whitelist_for_serdes(storage_field_names={"node_handle": "solid_handle"})
class StepHandle(NamedTuple("_StepHandle", [("node_handle", NodeHandle), ("key", str)])):
    """A reference to an ExecutionStep that was determined statically."""

    def __new__(cls, node_handle: NodeHandle, key: Optional[str] = None):
        return super().__new__(
            cls,
            node_handle=check.inst_param(node_handle, "node_handle", NodeHandle),
            # mypy can't tell that if default is set, this is guaranteed to be a str
            key=cast("str", check.opt_str_param(key, "key", default=str(node_handle))),
        )

    def to_key(self) -> str:
        return self.key

    @staticmethod
    def parse_from_key(
        string: str,
    ) -> Union["StepHandle", "ResolvedFromDynamicStepHandle", "UnresolvedStepHandle"]:
        unresolved_match = re.match(r"(.*)\[\?\]", string)
        if unresolved_match:
            return UnresolvedStepHandle(NodeHandle.from_string(unresolved_match.group(1)))

        resolved_match = re.match(r"(.*)\[(.*)\]", string)
        if resolved_match:
            return ResolvedFromDynamicStepHandle(
                NodeHandle.from_string(resolved_match.group(1)), resolved_match.group(2)
            )

        return StepHandle(NodeHandle.from_string(string))


# Serialize node_handle -> solid_handle for backcompat
@whitelist_for_serdes(storage_field_names={"node_handle": "solid_handle"})
class UnresolvedStepHandle(NamedTuple("_UnresolvedStepHandle", [("node_handle", NodeHandle)])):
    """A reference to an UnresolvedMappedExecutionStep in an execution."""

    def __new__(cls, node_handle: NodeHandle):
        return super().__new__(
            cls,
            node_handle=check.inst_param(node_handle, "node_handle", NodeHandle),
        )

    def to_key(self) -> str:
        return f"{self.node_handle}[?]"

    def resolve(self, map_key) -> "ResolvedFromDynamicStepHandle":
        return ResolvedFromDynamicStepHandle(self.node_handle, map_key)


# Serialize node_handle -> solid_handle for backcompat
@whitelist_for_serdes(storage_field_names={"node_handle": "solid_handle"})
class ResolvedFromDynamicStepHandle(
    NamedTuple(
        "_ResolvedFromDynamicStepHandle",
        [("node_handle", NodeHandle), ("mapping_key", str), ("key", str)],
    )
):
    """A reference to an ExecutionStep that came from resolving an UnresolvedMappedExecutionStep
    (and associated UnresolvedStepHandle) downstream of a dynamic output after it has
    completed successfully.
    """

    def __new__(cls, node_handle: NodeHandle, mapping_key: str, key: Optional[str] = None):
        return super().__new__(
            cls,
            node_handle=check.inst_param(node_handle, "node_handle", NodeHandle),
            mapping_key=check.str_param(mapping_key, "mapping_key"),
            # mypy can't tell that if default is set, this is guaranteed to be a str
            key=cast(
                "str",
                check.opt_str_param(key, "key", default=f"{node_handle}[{mapping_key}]"),
            ),
        )

    def to_key(self) -> str:
        return self.key

    @property
    def unresolved_form(self) -> UnresolvedStepHandle:
        return UnresolvedStepHandle(node_handle=self.node_handle)



from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Literal, TypedDict, Union, get_args

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.metadata.table import (
    TableColumn as TableColumn,
    TableColumnConstraints as TableColumnConstraints,
    TableColumnDep as TableColumnDep,
    TableColumnLineage as TableColumnLineage,
    TableConstraints as TableConstraints,
    TableRecord as TableRecord,
    TableSchema as TableSchema,
)

if TYPE_CHECKING:
    from dagster._core.definitions.metadata import MetadataValue

InferrableExternalMetadataValue = Union[
    int, float, str, Mapping[str, Any], Sequence[Any], bool, None
]


class ExternalMetadataValue(TypedDict):
    type: "ExternalMetadataType"
    raw_value: InferrableExternalMetadataValue


# Infer the type from the raw value on the orchestration end
EXTERNAL_METADATA_TYPE_INFER = "__infer__"

ExternalMetadataType = Literal[
    "__infer__",
    "text",
    "url",
    "path",
    "notebook",
    "json",
    "md",
    "float",
    "int",
    "bool",
    "dagster_run",
    "asset",
    "null",
    "table",
    "table_schema",
    "table_column_lineage",
    "timestamp",
]
EXTERNAL_METADATA_VALUE_KEYS = frozenset(ExternalMetadataValue.__annotations__.keys())
EXTERNAL_METADATA_TYPES = frozenset(get_args(ExternalMetadataType))


def metadata_map_from_external(
    metadata: Mapping[str, ExternalMetadataValue],
) -> Mapping[str, "MetadataValue"]:
    return {k: metadata_value_from_external(v["raw_value"], v["type"]) for k, v in metadata.items()}


def metadata_value_from_external(
    value: Any, metadata_type: ExternalMetadataType
) -> "MetadataValue":
    from dagster._core.definitions.metadata import MetadataValue, normalize_metadata_value

    if metadata_type == EXTERNAL_METADATA_TYPE_INFER:
        return normalize_metadata_value(value)
    elif metadata_type == "text":
        return MetadataValue.text(value)
    elif metadata_type == "url":
        return MetadataValue.url(value)
    elif metadata_type == "path":
        return MetadataValue.path(value)
    elif metadata_type == "notebook":
        return MetadataValue.notebook(value)
    elif metadata_type == "json":
        return MetadataValue.json(value)
    elif metadata_type == "md":
        return MetadataValue.md(value)
    elif metadata_type == "float":
        return MetadataValue.float(value)
    elif metadata_type == "int":
        return MetadataValue.int(value)
    elif metadata_type == "bool":
        return MetadataValue.bool(value)
    elif metadata_type == "dagster_run":
        return MetadataValue.dagster_run(value)
    elif metadata_type == "asset":
        return MetadataValue.asset(AssetKey.from_user_string(value))
    elif metadata_type == "table":
        value = check.mapping_param(value, "table_value", key_type=str)
        return MetadataValue.table(
            records=[TableRecord(record) for record in value["records"]],
            schema=TableSchema(
                columns=[
                    TableColumn(
                        name=column["name"],
                        type=column["type"],
                        description=column.get("description"),
                        tags=column.get("tags"),
                        constraints=TableColumnConstraints(**column["constraints"])
                        if column.get("constraints")
                        else None,
                    )
                    for column in value["schema"]
                ]
            ),
        )
    elif metadata_type == "table_schema":
        value = check.mapping_param(value, "table_schema_value", key_type=str)
        return MetadataValue.table_schema(
            schema=TableSchema(
                columns=[
                    TableColumn(
                        name=column["name"],
                        type=column["type"],
                        description=column.get("description"),
                        tags=column.get("tags"),
                        constraints=TableColumnConstraints(**column["constraints"])
                        if column.get("constraints")
                        else None,
                    )
                    for column in value["columns"]
                ]
            )
        )
    elif metadata_type == "table_column_lineage":
        value = check.mapping_param(value, "table_column_value", key_type=str)
        return MetadataValue.column_lineage(
            lineage=TableColumnLineage(
                deps_by_column={
                    column: [TableColumnDep(**dep) for dep in deps]
                    for column, deps in value["deps_by_column"].items()
                }
            )
        )
    elif metadata_type == "timestamp":
        return MetadataValue.timestamp(float(check.numeric_param(value, "timestamp")))
    elif metadata_type == "null":
        return MetadataValue.null()
    else:
        check.failed(f"Unexpected metadata type {metadata_type}")



from pathlib import Path
from typing import Optional

import click
from dagster_shared.serdes.objects import PluginObjectKey

from dagster.components.component_scaffolding import scaffold_object
from dagster.components.core.package_entry import load_package_object


@click.group(name="scaffold")
def scaffold_cli() -> None:
    """Commands for scaffolding Dagster components and related entities."""


@scaffold_cli.command(name="object")
@click.argument("typename", type=str)
@click.argument("path", type=Path)
@click.option(
    "--json-params", type=str, default=None, help="JSON string containing scaffold parameters"
)
@click.option(
    "--scaffold-format",
    type=click.Choice(["yaml", "python"], case_sensitive=False),
    help="Format of the component configuration (yaml or python)",
)
@click.option("--project-root", type=Path)
def scaffold_object_command(
    typename: str,
    path: Path,
    json_params: Optional[str],
    scaffold_format: str,
    project_root: Optional[Path],
) -> None:
    scaffold_object_command_impl(typename, path, json_params, scaffold_format, project_root)


def scaffold_object_command_impl(
    typename: str,
    path: Path,
    json_params: Optional[str],
    scaffold_format: str,
    project_root: Optional[Path],
) -> None:
    key = PluginObjectKey.from_typename(typename)
    obj = load_package_object(key)

    scaffold_object(
        path,
        obj,
        typename,
        json_params,
        scaffold_format,
        project_root,
    )



import logging
from collections.abc import Mapping, Sequence
from typing import Any, NamedTuple, Optional

from typing_extensions import Self

from dagster import (
    DagsterEvent,
    IntSource,
    String,
    _check as check,
)
from dagster._builtins import Bool
from dagster._config import Array, Field, Noneable, ScalarUnion, Shape
from dagster._config.config_schema import UserConfigSchema
from dagster._core.instance import T_DagsterInstance
from dagster._core.run_coordinator.base import RunCoordinator, SubmitRunContext
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._serdes import ConfigurableClass, ConfigurableClassData


class RunQueueConfig(
    NamedTuple(
        "_RunQueueConfig",
        [
            ("max_concurrent_runs", int),
            ("tag_concurrency_limits", Sequence[Mapping[str, Any]]),
            ("max_user_code_failure_retries", int),
            ("user_code_failure_retry_delay", int),
            ("should_block_op_concurrency_limited_runs", bool),
            ("op_concurrency_slot_buffer", int),
            ("explicitly_enabled_concurrency_run_blocking", bool),
        ],
    )
):
    def __new__(
        cls,
        max_concurrent_runs: int,
        tag_concurrency_limits: Optional[Sequence[Mapping[str, Any]]],
        max_user_code_failure_retries: int = 0,
        user_code_failure_retry_delay: int = 60,
        should_block_op_concurrency_limited_runs: bool = True,
        op_concurrency_slot_buffer: int = 0,
        explicitly_enabled_concurrency_run_blocking: bool = False,
    ):
        return super().__new__(
            cls,
            check.int_param(max_concurrent_runs, "max_concurrent_runs"),
            check.opt_sequence_param(tag_concurrency_limits, "tag_concurrency_limits"),
            check.int_param(max_user_code_failure_retries, "max_user_code_failure_retries"),
            check.int_param(user_code_failure_retry_delay, "user_code_failure_retry_delay"),
            check.bool_param(
                should_block_op_concurrency_limited_runs, "should_block_op_concurrency_limited_runs"
            ),
            check.int_param(op_concurrency_slot_buffer, "op_concurrency_slot_buffer"),
            check.bool_param(
                explicitly_enabled_concurrency_run_blocking,
                "explicitly_enabled_concurrency_run_blocking",
            ),
        )

    def with_concurrency_settings(
        self, concurrency_settings: Mapping[str, Any]
    ) -> "RunQueueConfig":
        run_settings = concurrency_settings.get("runs", {})
        pool_settings = concurrency_settings.get("pools", {})
        return RunQueueConfig(
            max_concurrent_runs=run_settings.get("max_concurrent_runs", self.max_concurrent_runs),
            tag_concurrency_limits=run_settings.get(
                "tag_concurrency_limits", self.tag_concurrency_limits
            ),
            max_user_code_failure_retries=self.max_user_code_failure_retries,
            user_code_failure_retry_delay=self.user_code_failure_retry_delay,
            should_block_op_concurrency_limited_runs=self.should_block_op_concurrency_limited_runs,
            op_concurrency_slot_buffer=pool_settings.get(
                "op_granularity_run_buffer", self.op_concurrency_slot_buffer
            ),
            explicitly_enabled_concurrency_run_blocking=self.explicitly_enabled_concurrency_run_blocking,
        )


class QueuedRunCoordinator(RunCoordinator[T_DagsterInstance], ConfigurableClass):
    """Enqueues runs via the run storage, to be deqeueued by the Dagster Daemon process. Requires
    the Dagster Daemon process to be alive in order for runs to be launched.
    """

    def __init__(
        self,
        max_concurrent_runs: Optional[int] = None,
        tag_concurrency_limits: Optional[Sequence[Mapping[str, Any]]] = None,
        dequeue_interval_seconds: Optional[int] = None,
        dequeue_use_threads: Optional[bool] = None,
        dequeue_num_workers: Optional[int] = None,
        max_user_code_failure_retries: Optional[int] = None,
        user_code_failure_retry_delay: Optional[int] = None,
        block_op_concurrency_limited_runs: Optional[Mapping[str, Any]] = None,
        inst_data: Optional[ConfigurableClassData] = None,
    ):
        self._inst_data: Optional[ConfigurableClassData] = check.opt_inst_param(
            inst_data, "inst_data", ConfigurableClassData
        )
        self._max_concurrent_runs: int = check.opt_int_param(
            max_concurrent_runs, "max_concurrent_runs", 10
        )
        check.invariant(
            self._max_concurrent_runs >= -1,
            "Negative values other than -1 (which disables the limit) for max_concurrent_runs"
            " are disallowed.",
        )
        self._tag_concurrency_limits: Sequence[Mapping[str, Any]] = check.opt_list_param(
            tag_concurrency_limits,
            "tag_concurrency_limits",
        )
        self._dequeue_interval_seconds: int = check.opt_int_param(
            dequeue_interval_seconds, "dequeue_interval_seconds", 5
        )
        self._dequeue_use_threads: bool = check.opt_bool_param(
            dequeue_use_threads, "dequeue_use_threads", False
        )
        self._dequeue_num_workers: Optional[int] = check.opt_int_param(
            dequeue_num_workers, "dequeue_num_workers"
        )
        self._max_user_code_failure_retries: int = check.opt_int_param(
            max_user_code_failure_retries, "max_user_code_failure_retries", 0
        )
        self._user_code_failure_retry_delay: int = check.opt_int_param(
            user_code_failure_retry_delay, "user_code_failure_retry_delay", 60
        )
        self._should_block_op_concurrency_limited_runs: bool = bool(
            not block_op_concurrency_limited_runs
            or block_op_concurrency_limited_runs.get("enabled", True)
        )
        self._explicitly_enabled_concurrency_run_blocking: bool = bool(
            block_op_concurrency_limited_runs and block_op_concurrency_limited_runs.get("enabled")
        )
        self._op_concurrency_slot_buffer: int = (
            block_op_concurrency_limited_runs.get("op_concurrency_slot_buffer", 0)
            if block_op_concurrency_limited_runs
            else 0
        )
        if self._op_concurrency_slot_buffer:
            check.invariant(
                self._should_block_op_concurrency_limited_runs,
                "op_concurrency_slot_buffer can only be set if block_op_concurrency_limited_runs "
                "is enabled",
            )
        self._logger = logging.getLogger("dagster.run_coordinator.queued_run_coordinator")
        super().__init__()

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    def get_run_queue_config(self) -> RunQueueConfig:
        return RunQueueConfig(
            max_concurrent_runs=self._max_concurrent_runs,
            tag_concurrency_limits=self._tag_concurrency_limits,
            max_user_code_failure_retries=self._max_user_code_failure_retries,
            user_code_failure_retry_delay=self._user_code_failure_retry_delay,
            should_block_op_concurrency_limited_runs=self._should_block_op_concurrency_limited_runs,
            op_concurrency_slot_buffer=self._op_concurrency_slot_buffer,
            explicitly_enabled_concurrency_run_blocking=self._explicitly_enabled_concurrency_run_blocking,
        )

    @property
    def dequeue_interval_seconds(self) -> int:
        return self._dequeue_interval_seconds

    @property
    def dequeue_use_threads(self) -> bool:
        return self._dequeue_use_threads

    @property
    def dequeue_num_workers(self) -> Optional[int]:
        return self._dequeue_num_workers

    @property
    def should_block_op_concurrency_limited_runs(self) -> bool:
        return self._should_block_op_concurrency_limited_runs

    @property
    def op_concurrency_slot_buffer(self) -> int:
        return self._op_concurrency_slot_buffer

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {
            "max_concurrent_runs": Field(
                config=IntSource,
                is_required=False,
                description=(
                    "The maximum number of runs that are allowed to be in progress at once."
                    " Defaults to 10. Set to -1 to disable the limit. Set to 0 to stop any runs"
                    " from launching. Any other negative values are disallowed."
                ),
            ),
            "tag_concurrency_limits": Field(
                config=Noneable(
                    Array(
                        Shape(
                            {
                                "key": String,
                                "value": Field(
                                    ScalarUnion(
                                        scalar_type=String,
                                        non_scalar_schema=Shape({"applyLimitPerUniqueValue": Bool}),
                                    ),
                                    is_required=False,
                                ),
                                "limit": Field(int),
                            }
                        )
                    )
                ),
                is_required=False,
                description=(
                    "A set of limits that are applied to runs with particular tags. If a value is"
                    " set, the limit is applied to only that key-value pair. If no value is set,"
                    " the limit is applied across all values of that key. If the value is set to a"
                    " dict with `applyLimitPerUniqueValue: true`, the limit will apply to the"
                    " number of unique values for that key."
                ),
            ),
            "dequeue_interval_seconds": Field(
                config=IntSource,
                is_required=False,
                description=(
                    "The interval in seconds at which the Dagster Daemon "
                    "should periodically check the run queue for new runs to launch."
                ),
            ),
            "dequeue_use_threads": Field(
                config=bool,
                is_required=False,
                description=(
                    "Whether or not to use threads for concurrency when launching dequeued runs."
                ),
            ),
            "dequeue_num_workers": Field(
                config=IntSource,
                is_required=False,
                description=(
                    "If dequeue_use_threads is true, limit the number of concurrent worker threads."
                ),
            ),
            "max_user_code_failure_retries": Field(
                config=IntSource,
                is_required=False,
                default_value=0,
                description=(
                    "If there is an error reaching a Dagster gRPC server while dequeuing the run,"
                    " how many times to retry the dequeue before failing it. The only run launcher"
                    " that requires the gRPC server to be running is the DefaultRunLauncher, so"
                    " setting this will have no effect unless that run launcher is being used."
                ),
            ),
            "user_code_failure_retry_delay": Field(
                config=IntSource,
                is_required=False,
                default_value=60,
                description=(
                    "If there is an error reaching a Dagster gRPC server while dequeuing the run,"
                    " how long to wait before retrying any runs from that same code location. The"
                    " only run launcher that requires the gRPC server to be running is the"
                    " DefaultRunLauncher, so setting this will have no effect unless that run"
                    " launcher is being used."
                ),
            ),
            "block_op_concurrency_limited_runs": Field(
                {
                    "enabled": Field(Bool, is_required=False),
                    "op_concurrency_slot_buffer": Field(
                        int,
                        is_required=False,
                        description=(
                            "Determines whether or not a run will be dequeued if it consists of ops that "
                            "will all be initially blocked waiting for global op concurrency slots to be "
                            "free."
                        ),
                    ),
                }
            ),
        }

    @classmethod
    def from_config_value(
        cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]
    ) -> Self:
        return cls(
            inst_data=inst_data,
            max_concurrent_runs=config_value.get("max_concurrent_runs"),
            tag_concurrency_limits=config_value.get("tag_concurrency_limits"),
            dequeue_interval_seconds=config_value.get("dequeue_interval_seconds"),
            dequeue_use_threads=config_value.get("dequeue_use_threads"),
            dequeue_num_workers=config_value.get("dequeue_num_workers"),
            max_user_code_failure_retries=config_value.get("max_user_code_failure_retries"),
            user_code_failure_retry_delay=config_value.get("user_code_failure_retry_delay"),
            block_op_concurrency_limited_runs=config_value.get("block_op_concurrency_limited_runs"),
        )

    def submit_run(self, context: SubmitRunContext) -> DagsterRun:
        dagster_run = context.dagster_run

        if dagster_run.status == DagsterRunStatus.NOT_STARTED:
            enqueued_event = DagsterEvent.job_enqueue(dagster_run)
            self._instance.report_dagster_event(enqueued_event, run_id=dagster_run.run_id)
        else:
            # the run was already submitted, this is a no-op
            self._logger.warning(
                f"submit_run called for run {dagster_run.run_id} with status "
                f"{dagster_run.status.value}, skipping enqueue."
            )

        run = self._instance.get_run_by_id(dagster_run.run_id)
        if run is None:
            check.failed(f"Failed to reload run {dagster_run.run_id}")
        assert run
        return run

    def cancel_run(self, run_id: str) -> bool:
        run = self._instance.get_run_by_id(run_id)
        if not run:
            return False
        # NOTE: possible race condition if the dequeuer acts on this run at the same time
        # https://github.com/dagster-io/dagster/issues/3323
        if run.status == DagsterRunStatus.QUEUED:
            self._instance.report_run_canceling(
                run,
                message="Canceling run from the queue.",
            )
            self._instance.report_run_canceled(run)
            return True
        else:
            return self._instance.run_launcher.terminate(run_id)



import shutil
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from dagster._annotations import preview, public
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.core_models import ResolvedAssetSpec
from dagster.components.resolved.model import Model

if TYPE_CHECKING:
    from dagster._core.definitions.definitions_class import Definitions


@dataclass
class PipesSubprocessScript(Resolvable):
    path: str
    assets: Sequence[ResolvedAssetSpec]


@public
@preview(emit_runtime_warning=False)
@dataclass
class PipesSubprocessScriptCollectionComponent(Component, Resolvable):
    """Assets that wrap Python scripts executed with Dagster's PipesSubprocessClient."""

    scripts: Sequence[PipesSubprocessScript] = field(
        metadata=dict(description="Scripts to be executed.")
    )

    @cached_property
    def specs_by_path(self) -> Mapping[str, Sequence[AssetSpec]]:
        return {script.path: script.assets for script in self.scripts}

    def build_defs(self, context: ComponentLoadContext) -> "Definitions":
        from dagster._core.definitions.definitions_class import Definitions

        return Definitions(
            assets=[
                _create_asset_def(context.path / path, specs)
                for path, specs in self.specs_by_path.items()
            ],
        )


def _create_asset_def(path: Path, specs: Sequence[AssetSpec]) -> AssetsDefinition:
    from dagster._core.pipes.subprocess import PipesSubprocessClient

    # TODO: allow name paraeterization
    @multi_asset(specs=specs, name=f"script_{path.stem}")
    def _asset(context: AssetExecutionContext):
        cmd = [shutil.which("python"), path]
        return PipesSubprocessClient().run(command=cmd, context=context).get_results()

    return _asset


@public
@preview(emit_runtime_warning=False)
class PipesSubprocessScriptComponent(Component, Resolvable, Model):
    path: str
    assets: Sequence[ResolvedAssetSpec]

    def build_defs(self, context: ComponentLoadContext) -> "Definitions":
        from dagster._core.definitions.definitions_class import Definitions

        return Definitions(assets=[_create_asset_def(context.path / self.path, self.assets)])



import asyncio
from collections.abc import Mapping, Sequence
from typing import Optional, Union

from dagster_shared.record import replace
from dagster_shared.serdes import whitelist_for_serdes

from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.metadata import MetadataMapping
from dagster._core.definitions.metadata.metadata_value import FloatMetadataValue, IntMetadataValue
from dagster._record import copy, record


@record
class SinceConditionData:
    """Convenience class for manipulating metadata relevant to historical SinceCondition evaluations.
    Tracks the previous evaluation id and timestamp of the last evaluation where the trigger condition
    and reset conditions were true.
    """

    trigger_evaluation_id: Optional[int]
    trigger_timestamp: Optional[float]
    reset_evaluation_id: Optional[int]
    reset_timestamp: Optional[float]

    @staticmethod
    def from_metadata(metadata: Optional[MetadataMapping]) -> "SinceConditionData":
        def _get_int(key: str) -> Optional[int]:
            metadata_val = metadata.get(key, None) if metadata else None
            return metadata_val.value if isinstance(metadata_val, IntMetadataValue) else None

        def _get_float(key: str) -> Optional[float]:
            metadata_val = metadata.get(key, None) if metadata else None
            return metadata_val.value if isinstance(metadata_val, FloatMetadataValue) else None

        return SinceConditionData(
            trigger_evaluation_id=_get_int("trigger_evaluation_id"),
            trigger_timestamp=_get_float("trigger_timestamp"),
            reset_evaluation_id=_get_int("reset_evaluation_id"),
            reset_timestamp=_get_float("reset_timestamp"),
        )

    def to_metadata(self) -> Mapping[str, Union[IntMetadataValue, FloatMetadataValue]]:
        return dict(
            trigger_evaluation_id=IntMetadataValue(self.trigger_evaluation_id),
            trigger_timestamp=FloatMetadataValue(self.trigger_timestamp),
            reset_evaluation_id=IntMetadataValue(self.reset_evaluation_id),
            reset_timestamp=FloatMetadataValue(self.reset_timestamp),
        )

    def update(
        self,
        evaluation_id: int,
        timestamp: float,
        trigger_result: AutomationResult,
        reset_result: AutomationResult,
    ) -> "SinceConditionData":
        updated = self
        if not trigger_result.true_subset.is_empty:
            updated = replace(
                updated, trigger_evaluation_id=evaluation_id, trigger_timestamp=timestamp
            )
        if not reset_result.true_subset.is_empty:
            updated = replace(updated, reset_evaluation_id=evaluation_id, reset_timestamp=timestamp)
        return updated


@whitelist_for_serdes
@record
class SinceCondition(BuiltinAutomationCondition[T_EntityKey]):
    trigger_condition: AutomationCondition[T_EntityKey]
    reset_condition: AutomationCondition[T_EntityKey]

    @property
    def name(self) -> str:
        return "SINCE"

    @property
    def children(self) -> Sequence[AutomationCondition[T_EntityKey]]:
        return [self.trigger_condition, self.reset_condition]

    async def evaluate(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        # must evaluate child condition over the entire subset to avoid missing state transitions
        child_candidate_subset = context.asset_graph_view.get_full_subset(key=context.key)

        # compute result for trigger and reset conditions
        trigger_result, reset_result = await asyncio.gather(
            *[
                context.for_child_condition(
                    self.trigger_condition,
                    child_indices=[0],
                    candidate_subset=child_candidate_subset,
                ).evaluate_async(),
                context.for_child_condition(
                    self.reset_condition,
                    child_indices=[1],
                    candidate_subset=child_candidate_subset,
                ).evaluate_async(),
            ]
        )

        # take the previous subset that this was true for
        true_subset = context.previous_true_subset or context.get_empty_subset()

        # add in any newly true trigger asset partitions
        true_subset = true_subset.compute_union(trigger_result.true_subset)
        # remove any newly true reset asset partitions
        true_subset = true_subset.compute_difference(reset_result.true_subset)

        # if anything changed since the previous evaluation, update the metadata
        condition_data = SinceConditionData.from_metadata(context.previous_metadata).update(
            context.evaluation_id,
            context.evaluation_time.timestamp(),
            trigger_result=trigger_result,
            reset_result=reset_result,
        )

        return AutomationResult(
            context=context,
            true_subset=true_subset,
            child_results=[trigger_result, reset_result],
            metadata=condition_data.to_metadata(),
        )

    def replace(
        self, old: Union[AutomationCondition, str], new: AutomationCondition
    ) -> AutomationCondition:
        """Replaces all instances of ``old`` across any sub-conditions with ``new``.

        If ``old`` is a string, then conditions with a label matching
        that string will be replaced.

        Args:
            old (Union[AutomationCondition, str]): The condition to replace.
            new (AutomationCondition): The condition to replace with.
        """
        return (
            new
            if old in [self, self.get_label()]
            else copy(
                self,
                trigger_condition=self.trigger_condition.replace(old, new),
                reset_condition=self.reset_condition.replace(old, new),
            )
        )



from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Generic, Literal, Optional, Union

from pydantic import BaseModel
from typing_extensions import TypeAlias, TypeVar

from dagster import _check as check
from dagster._annotations import preview, public

# Constant for object attribute name
SCAFFOLDER_CLS_ATTRIBUTE: str = "__scaffolder_cls__"


class NoParams(BaseModel): ...


# Type variable for generic class handling
T = TypeVar("T")
TModel = TypeVar("TModel", bound=BaseModel, default=NoParams)


@public
@preview(emit_runtime_warning=False)
def scaffold_with(
    scaffolder_cls: Union[type["Scaffolder[Any]"], "ScaffolderUnavailableReason"],
) -> Callable[[T], T]:
    """A decorator that declares what Scaffolder is used to scaffold the artifact.

    Args:
        scaffolder_cls: A class that inherits from Scaffolder

    Returns:
        Decorator function that enhances the target class
    """
    from dagster.components.core.package_entry import PACKAGE_ENTRY_ATTR

    def decorator(obj: T) -> T:
        # Store the scaffolder class as an attribute using the constant
        setattr(obj, SCAFFOLDER_CLS_ATTRIBUTE, scaffolder_cls)
        # All scaffoldable objects are library objects
        setattr(obj, PACKAGE_ENTRY_ATTR, True)
        return obj

    return decorator


def has_scaffolder(obj: object) -> bool:
    """Determines if an object has been decorated with `@scaffold_with`.

    Args:
        obj: The object to check

    Returns:
        True if the object has a Scaffolder attached, False otherwise
    """
    return hasattr(obj, SCAFFOLDER_CLS_ATTRIBUTE)


def get_scaffolder(obj: object) -> Union["Scaffolder[Any]", "ScaffolderUnavailableReason"]:
    """Retrieves the scaffolder class attached to the decorated object.

    Args:
        obj: The object to inspect

    Returns:
        The scaffolder class attached to the decorated object. Raises CheckError if the object is not decorated with @scaffold_with.
    """
    check.param_invariant(
        has_scaffolder(obj), "obj", "Object must be decorated with @scaffold_with"
    )
    attr = getattr(obj, SCAFFOLDER_CLS_ATTRIBUTE)
    return attr if isinstance(attr, ScaffolderUnavailableReason) else attr()


@dataclass
class ScaffolderUnavailableReason:
    message: str


ScaffoldFormatOptions: TypeAlias = Literal["yaml", "python"]


@public
@preview(emit_runtime_warning=False)
@dataclass
class ScaffoldRequest(Generic[TModel]):
    """The details about the current scaffolding operation."""

    # fully qualified class name of the decorated object
    type_name: str
    # target path for the scaffold request. Typically used to construct absolute paths
    target_path: Path
    # yaml or python
    scaffold_format: ScaffoldFormatOptions
    # the root of the dg project
    project_root: Optional[Path]
    # optional params for scaffolding
    params: TModel


@public
@preview(emit_runtime_warning=False)
class Scaffolder(Generic[TModel]):
    """Handles scaffolding its associated scaffold target."""

    @classmethod
    def get_scaffold_params(cls) -> type[TModel]:
        return NoParams  # type: ignore

    @abstractmethod
    def scaffold(self, request: ScaffoldRequest[TModel]) -> None:
        """Scaffold the target with the given request.

        Args:
            request: The scaffold request containing type name, target path, format, project root and params
        """
        ...



from pathlib import Path
from typing import Optional

from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.core.defs_module import DagsterDefsComponent
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.model import Model

############
# BACKCOMPAT
############


class DefinitionsComponent(Component, Model, Resolvable):
    """An arbitrary set of dagster definitions."""

    path: Optional[str]

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        component = DagsterDefsComponent(Path(self.path) if self.path else context.path)
        return component.build_defs(context)



import os
from pathlib import Path
from typing import Optional

import click
from dagster_shared.scaffold import scaffold_subtree

from dagster.version import __version__ as dagster_version

PROJECT_NAME_PLACEHOLDER = "PROJECT_NAME_PLACEHOLDER"


def generate_repository(path: str):
    REPO_NAME_PLACEHOLDER = "REPO_NAME_PLACEHOLDER"

    click.echo(f"Creating a Dagster repository at {path}.")

    #
    generate_project(
        path=path,
        excludes=None,
        name_placeholder=REPO_NAME_PLACEHOLDER,
        templates_path=os.path.join(os.path.dirname(__file__), "templates", REPO_NAME_PLACEHOLDER),
    )

    click.echo(f"Generated files for Dagster repository in {path}.")


def generate_project(
    path: str,
    excludes: Optional[list[str]] = None,
    name_placeholder: str = PROJECT_NAME_PLACEHOLDER,
    templates_path: str = PROJECT_NAME_PLACEHOLDER,
):
    """Renders templates for Dagster project."""
    click.echo(f"Creating a Dagster project at {path}.")

    project_template_path = os.path.join(os.path.dirname(__file__), "templates", templates_path)

    return scaffold_subtree(
        path=Path(path),
        project_template_path=Path(project_template_path),
        excludes=excludes,
        name_placeholder=name_placeholder,
        templates_path=templates_path,
        dagster_version=dagster_version,
    )

    click.echo(f"Generated files for Dagster project in {path}.")



from typing import Optional

from dagster_shared import record
from dagster_shared.serdes import whitelist_for_serdes

import dagster._check as check
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.loader import LoadingContext
from dagster._core.remote_representation.external_data import PartitionsSnap
from dagster._core.storage.dagster_run import RunRecord
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.storage.partition_status_cache import get_partition_subsets
from dagster._streamline.asset_health import AssetHealthStatus


@whitelist_for_serdes
@record.record
class AssetMaterializationHealthState:
    """For tracking the materialization health of an asset, we only care about the most recent
    completed materialization attempt for each asset/partition. This record keeps track of the
    assets/partitions that have ever been successfully materialized and those that are currently in
    a failed state. From this information we can derive the subset that is currently in a successfully
    materialized state.

    If an asset/partition is currently being materialized, it will not move to a new state until after
    the materialization attempt is complete.

    In the future, we may want to expand this to track the last N materialization successes/failures for
    each asset. We could also maintain a list of in progress materializations, but that requires streamline to be
    better able to handle runs being deleted.

    materialized_subset: The subset of the asset that has ever been successfully materialized.
    failed_subset: The subset of the asset that is currently in a failed state.
    partitions_snap: The partitions definition for the asset. None if it is not a partitioned asset.
    latest_terminal_run_id: The id of the latest run with a successful or failed materialization event for the asset.
    """

    materialized_subset: SerializableEntitySubset[AssetKey]
    failed_subset: SerializableEntitySubset[AssetKey]
    partitions_snap: Optional[PartitionsSnap]
    latest_terminal_run_id: Optional[str]

    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        if self.partitions_snap is None:
            return None
        return self.partitions_snap.get_partitions_definition()

    @property
    def currently_materialized_subset(self) -> SerializableEntitySubset[AssetKey]:
        """The subset of the asset that is currently in a successfully materialized state."""
        return self.materialized_subset.compute_difference(self.failed_subset)

    @property
    def health_status(self) -> AssetHealthStatus:
        if self.materialized_subset.is_empty and self.failed_subset.is_empty:
            return AssetHealthStatus.UNKNOWN
        elif not self.failed_subset.is_empty:
            return AssetHealthStatus.DEGRADED
        else:
            return AssetHealthStatus.HEALTHY

    @classmethod
    async def compute_for_asset(
        cls,
        asset_key: AssetKey,
        partitions_def: Optional[PartitionsDefinition],
        loading_context: LoadingContext,
    ) -> "AssetMaterializationHealthState":
        """Creates an AssetMaterializationHealthState for the given asset. Requires fetching the AssetRecord
        and potentially the latest run from the DB, or regenerating the partition status cache.
        """
        asset_record = await AssetRecord.gen(loading_context, asset_key)

        if partitions_def is not None:
            (
                materialized_partition_subset,
                failed_partition_subset,
                _,
            ) = get_partition_subsets(
                loading_context.instance,
                loading_context,
                asset_key,
                loading_context.instance,
                partitions_def,
            )

            if materialized_partition_subset is None or failed_partition_subset is None:
                check.failed("Expected partitions subset for a partitioned asset")

            last_run_id = None
            if asset_record is not None:
                entry = asset_record.asset_entry
                latest_record = max(
                    [
                        entry.last_materialization_record,
                        entry.last_failed_to_materialize_record,
                    ],
                    key=lambda record: -1 if record is None else record.storage_id,
                )
                last_run_id = latest_record.run_id if latest_record else None

            return cls(
                materialized_subset=SerializableEntitySubset(
                    key=asset_key, value=materialized_partition_subset
                ),
                failed_subset=SerializableEntitySubset(
                    key=asset_key, value=failed_partition_subset
                ),
                partitions_snap=PartitionsSnap.from_def(partitions_def),
                latest_terminal_run_id=last_run_id,
            )

        if asset_record is None:
            return AssetMaterializationHealthState(
                materialized_subset=SerializableEntitySubset(key=asset_key, value=False),
                failed_subset=SerializableEntitySubset(key=asset_key, value=False),
                partitions_snap=None,
                latest_terminal_run_id=None,
            )

        asset_entry = asset_record.asset_entry
        if asset_entry.last_run_id is None:
            return AssetMaterializationHealthState(
                materialized_subset=SerializableEntitySubset(key=asset_key, value=False),
                failed_subset=SerializableEntitySubset(key=asset_key, value=False),
                partitions_snap=None,
                latest_terminal_run_id=None,
            )

        has_ever_materialized = asset_entry.last_materialization is not None
        (
            is_currently_failed,
            latest_terminal_run_id,
        ) = await _get_is_currently_failed_and_latest_terminal_run_id(loading_context, asset_record)

        return cls(
            materialized_subset=SerializableEntitySubset(
                key=asset_key, value=has_ever_materialized
            ),
            failed_subset=SerializableEntitySubset(key=asset_key, value=is_currently_failed),
            partitions_snap=None,
            latest_terminal_run_id=latest_terminal_run_id,
        )


async def _get_is_currently_failed_and_latest_terminal_run_id(
    loading_context: LoadingContext, asset_record: AssetRecord
) -> tuple[bool, Optional[str]]:
    """Determines if the asset is currently in a failed state. If we are storing failure events for the
    asset, this can be determined by looking at the AssetRecord. For assets where we are not storing failure
    events, we have to derive the failure state from the latest run record.

    Also returns the id of the latest run with a successful or failed materialization event for the asset.
    """
    asset_entry = asset_record.asset_entry
    if loading_context.instance.can_read_failure_events_for_asset(asset_record):
        latest_record = max(
            [
                asset_entry.last_materialization_record,
                asset_entry.last_failed_to_materialize_record,
            ],
            key=lambda record: -1 if record is None else record.storage_id,
        )
        return (
            latest_record.storage_id == asset_entry.last_failed_to_materialize_storage_id
            if latest_record
            else False,
            latest_record.run_id if latest_record else None,
        )

    # if failure events are not stored, we usually have to fetch the run record to check if the
    # asset is currently failed. However, if the latest run id is the same as the last materialization run id,
    # then we know the asset is in a successfully materialized state.
    if (
        asset_entry.last_materialization
        and asset_entry.last_run_id == asset_entry.last_materialization.run_id
    ):
        return False, asset_entry.last_materialization.run_id

    run_record = await RunRecord.gen(loading_context, check.not_none(asset_entry.last_run_id))
    if run_record is None or not run_record.dagster_run.is_finished:
        # the run is deleted or in progress. With the information we have available, we cannot know
        # if the asset is in a failed state prior to this run. Historically, we have resorted to
        # reporting the asset as materialized if it has ever been materialized, and otherwise report it
        # as not materialized.
        return (
            False,
            asset_entry.last_materialization.run_id if asset_entry.last_materialization else None,
        )

    run_end_time = check.not_none(run_record.end_time)
    if (
        asset_entry.last_materialization
        and asset_entry.last_materialization.timestamp > run_end_time
    ):
        # the latest materialization was reported manually
        return False, asset_entry.last_materialization.run_id

    # if the run failed, then report the asset as failed
    return run_record.dagster_run.is_failure, run_record.dagster_run.run_id



import inspect
from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

from dagster_shared.record import record
from dagster_shared.serdes.objects import PluginObjectKey
from dagster_shared.yaml_utils import parse_yamls_with_source_position
from dagster_shared.yaml_utils.source_position import SourcePosition
from pydantic import BaseModel, ConfigDict, TypeAdapter

import dagster._check as check
from dagster._annotations import preview, public
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.metadata.source_code import (
    CodeReferencesMetadataSet,
    CodeReferencesMetadataValue,
)
from dagster._core.definitions.module_loaders.load_defs_from_module import (
    load_definitions_from_module,
)
from dagster._core.definitions.module_loaders.utils import find_objects_in_module_of_types
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils.pydantic_yaml import (
    _parse_and_populate_model_with_annotated_errors,
    enrich_validation_errors_with_source_position,
)
from dagster.components.component.component import Component
from dagster.components.component.component_loader import is_component_loader
from dagster.components.core.context import ComponentLoadContext, use_component_load_context
from dagster.components.core.package_entry import load_package_object
from dagster.components.definitions import LazyDefinitions
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.core_models import AssetPostProcessor

T = TypeVar("T", bound=BaseModel)


class ComponentRequirementsModel(BaseModel):
    """Describes dependencies for a component to load."""

    env: Optional[list[str]] = None


class ComponentFileModel(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: str
    attributes: Optional[Mapping[str, Any]] = None
    requirements: Optional[ComponentRequirementsModel] = None


def _add_defs_yaml_code_reference_to_spec(
    component_yaml_path: Path,
    load_context: ComponentLoadContext,
    component: Component,
    source_position: SourcePosition,
    asset_spec: AssetSpec,
) -> AssetSpec:
    existing_references_meta = CodeReferencesMetadataSet.extract(asset_spec.metadata)

    references = (
        existing_references_meta.code_references.code_references
        if existing_references_meta.code_references
        else []
    )
    references_to_add = component.get_code_references_for_yaml(
        component_yaml_path, source_position, load_context
    )

    return asset_spec.merge_attributes(
        metadata={
            **CodeReferencesMetadataSet(
                code_references=CodeReferencesMetadataValue(
                    code_references=[
                        *references,
                        *references_to_add,
                    ],
                )
            ),
        }
    )


class CompositeYamlComponent(Component):
    def __init__(self, components: Sequence[Component], source_positions: Sequence[SourcePosition]):
        self.components = components
        self.source_positions = source_positions

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        component_yaml = check.not_none(_find_defs_or_component_yaml(context.path))

        return Definitions.merge(
            *(
                component.build_defs(context).map_asset_specs(
                    func=lambda spec: _add_defs_yaml_code_reference_to_spec(
                        component_yaml_path=component_yaml,
                        load_context=context,
                        component=component,
                        source_position=source_position,
                        asset_spec=spec,
                    )
                )
                for component, source_position in zip(self.components, self.source_positions)
            )
        )


class CompositeComponent(Component):
    def __init__(self, components: Sequence[Component]):
        self.components = components

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return Definitions.merge(*[component.build_defs(context) for component in self.components])


def get_component(context: ComponentLoadContext) -> Optional[Component]:
    """Attempts to load a component from the given context. Iterates through potential component
    type matches, prioritizing more specific types: YAML, Python, plain Dagster defs, and component
    folder.
    """
    # in priority order
    # yaml component
    if _find_defs_or_component_yaml(context.path):
        return load_yaml_component(context)
    # pythonic component
    elif (context.path / "component.py").exists():
        return load_pythonic_component(context)
    # defs
    elif (context.path / "definitions.py").exists():
        return DagsterDefsComponent(path=context.path / "definitions.py")
    elif context.path.suffix == ".py":
        return DagsterDefsComponent(path=context.path)
    # folder
    elif context.path.is_dir():
        children = find_components_from_context(context)
        if children:
            return DefsFolderComponent(
                path=context.path,
                children=children,
                asset_post_processors=None,
            )

    return None


@dataclass
class DefsFolderComponentYamlSchema(Resolvable):
    asset_post_processors: Optional[Sequence[AssetPostProcessor]] = None


@record
class ComponentPath:
    """Identifier for where a Component instance was defined:
    file_path: The Path to the file or directory.
    instance_key: The optional identifier to distinguish instances originating from the same file.
    """

    file_path: Path
    instance_key: Optional[Union[int, str]] = None


@public
@preview(emit_runtime_warning=False)
@dataclass
class DefsFolderComponent(Component):
    """A folder which may contain multiple submodules, each
    which define components.

    Optionally enables postprocessing to modify the Dagster definitions
    produced by submodules.
    """

    path: Path
    children: Mapping[Path, Component]
    asset_post_processors: Optional[Sequence[AssetPostProcessor]]

    @classmethod
    def get_model_cls(cls):
        return DefsFolderComponentYamlSchema.model()

    @classmethod
    def load(cls, attributes: Any, context: ComponentLoadContext) -> "DefsFolderComponent":
        # doing this funky thing because some of our attributes are resolved from the context,
        # so we split up resolving the yaml-defined attributes and the context-defined attributes,
        # meaning we manually invoke the resolution system here
        resolved_attributes = DefsFolderComponentYamlSchema.resolve_from_model(
            context.resolution_context.at_path("attributes"),
            attributes,
        )

        return DefsFolderComponent(
            path=context.path,
            children=find_components_from_context(context),
            asset_post_processors=resolved_attributes.asset_post_processors,
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        child_defs = []
        for path, child in self.children.items():
            sub_ctx = context.for_path(path)
            with use_component_load_context(sub_ctx):
                child_defs.append(child.build_defs(sub_ctx))
        defs = Definitions.merge(*child_defs)
        for post_processor in self.asset_post_processors or []:
            defs = post_processor(defs)
        return defs

    @classmethod
    def get(cls, context: ComponentLoadContext) -> "DefsFolderComponent":
        component = get_component(context)
        return check.inst(
            component,
            DefsFolderComponent,
            f"Expected DefsFolderComponent at {context.path}, got {component}.",
        )

    def iterate_components(self) -> Iterator[Component]:
        for _, component in self.iterate_path_component_pairs():
            yield component

    def iterate_path_component_pairs(self) -> Iterator[tuple[ComponentPath, Component]]:
        for path, component in self.children.items():
            yield ComponentPath(file_path=path), component

            if isinstance(component, DefsFolderComponent):
                yield from component.iterate_path_component_pairs()

            if isinstance(component, CompositeYamlComponent):
                for idx, inner_comp in enumerate(component.components):
                    yield ComponentPath(file_path=path, instance_key=idx), inner_comp


EXPLICITLY_IGNORED_GLOB_PATTERNS = [
    "__pycache__",
    ".*/",
]


def find_components_from_context(context: ComponentLoadContext) -> Mapping[Path, Component]:
    found = {}
    for subpath in context.path.iterdir():
        relative_subpath = subpath.relative_to(context.path)
        if any(relative_subpath.match(pattern) for pattern in EXPLICITLY_IGNORED_GLOB_PATTERNS):
            continue
        sub_ctx = context.for_path(subpath)
        with use_component_load_context(sub_ctx):
            component = get_component(sub_ctx)
            if component:
                found[subpath] = component
    return found


@dataclass
class DagsterDefsComponent(Component):
    """A Python module containing Dagster definitions. Used for implicit loading of
    Dagster definitions from Python files in the defs folder.
    """

    path: Path

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        module = context.load_defs_relative_python_module(self.path)

        def_objects = check.is_list(
            list(find_objects_in_module_of_types(module, Definitions)), Definitions
        )
        lazy_def_objects = check.is_list(
            list(find_objects_in_module_of_types(module, LazyDefinitions)), LazyDefinitions
        )

        if lazy_def_objects and def_objects:
            raise DagsterInvalidDefinitionError(
                f"Found both @definitions-decorated functions and Definitions objects in {self.path}. "
                "At most one may be specified per module."
            )

        if len(def_objects) == 1:
            return next(iter(def_objects))

        if len(def_objects) > 1:
            raise DagsterInvalidDefinitionError(
                f"Found multiple Definitions objects in {self.path}. At most one Definitions object "
                "may be specified per module."
            )

        if len(lazy_def_objects) == 1:
            lazy_def = next(iter(lazy_def_objects))
            return lazy_def(context)

        if len(lazy_def_objects) > 1:
            raise DagsterInvalidDefinitionError(
                f"Found multiple @definitions-decorated functions in {self.path}. At most one "
                "@definitions-decorated function may be specified per module."
            )

        return load_definitions_from_module(module)


def load_pythonic_component(context: ComponentLoadContext) -> Component:
    # backcompat for component.yaml
    component_def_path = context.path / "component.py"
    module = context.load_defs_relative_python_module(component_def_path)
    component_loaders = list(inspect.getmembers(module, is_component_loader))
    if len(component_loaders) == 0:
        raise DagsterInvalidDefinitionError("No component loaders found in module")
    elif len(component_loaders) == 1:
        _, component_loader = component_loaders[0]
        return component_loader(context)
    else:
        return CompositeComponent(
            [component_loader(context) for _, component_loader in component_loaders]
        )


def load_yaml_component(context: ComponentLoadContext) -> Component:
    # parse the yaml file
    component_def_path = check.not_none(_find_defs_or_component_yaml(context.path))
    return load_yaml_component_from_path(context, component_def_path)


def load_yaml_component_from_path(context: ComponentLoadContext, component_def_path: Path):
    source_trees = parse_yamls_with_source_position(
        component_def_path.read_text(), str(component_def_path)
    )
    components = []
    for source_tree in source_trees:
        component_file_model = _parse_and_populate_model_with_annotated_errors(
            cls=ComponentFileModel, obj_parse_root=source_tree, obj_key_path_prefix=[]
        )

        # find the component type
        type_str = context.normalize_component_type_str(component_file_model.type)
        key = PluginObjectKey.from_typename(type_str)
        obj = load_package_object(key)
        if not isinstance(obj, type) or not issubclass(obj, Component):
            raise DagsterInvalidDefinitionError(
                f"Component type {type_str} is of type {type(obj)}, but must be a subclass of dagster.Component"
            )

        model_cls = obj.get_model_cls()
        context = context.with_rendering_scope(
            obj.get_additional_scope(),
        ).with_source_position_tree(
            source_tree.source_position_tree,
        )

        # grab the attributes from the yaml file
        if model_cls is None:
            attributes = None
        elif source_tree:
            attributes_position_tree = source_tree.source_position_tree.children["attributes"]
            with enrich_validation_errors_with_source_position(
                attributes_position_tree, ["attributes"]
            ):
                attributes = TypeAdapter(model_cls).validate_python(component_file_model.attributes)
        else:
            attributes = TypeAdapter(model_cls).validate_python(component_file_model.attributes)

        components.append(obj.load(attributes, context))

    check.invariant(len(components) > 0, "No components found in YAML file")
    return CompositeYamlComponent(
        components, [source_tree.source_position_tree.position for source_tree in source_trees]
    )


# When we remove component.yaml, we can remove this function for just a defs.yaml check
def _find_defs_or_component_yaml(path: Path) -> Optional[Path]:
    # Check for defs.yaml has precedence, component.yaml is deprecated
    return next(
        (p for p in (path / "defs.yaml", path / "component.yaml") if p.exists()),
        None,
    )



import json
import logging
import sys
from collections.abc import Sequence
from pathlib import Path
from traceback import TracebackException
from typing import Any, Literal, Optional, Union

import click
from dagster_shared.error import SerializableErrorInfo, remove_system_frames_from_error
from dagster_shared.serdes.objects import PluginObjectKey
from dagster_shared.serdes.objects.definition_metadata import (
    DgAssetCheckMetadata,
    DgAssetMetadata,
    DgDefinitionMetadata,
    DgJobMetadata,
    DgResourceMetadata,
    DgScheduleMetadata,
    DgSensorMetadata,
)
from dagster_shared.serdes.objects.package_entry import PluginManifest
from dagster_shared.serdes.serdes import serialize_value
from pydantic import ConfigDict, TypeAdapter, create_model

from dagster._cli.utils import assert_no_remaining_opts, get_possibly_temporary_instance_for_cli
from dagster._cli.workspace.cli_target import (
    PythonPointerOpts,
    get_repository_python_origin_from_cli_opts,
    python_pointer_options,
)
from dagster._config.pythonic_config.resource import get_resource_type_name
from dagster._core.definitions.asset_job import is_reserved_asset_job_name
from dagster._utils.error import serializable_error_info_from_exc_info
from dagster._utils.hosted_user_process import recon_repository_from_origin
from dagster.components.component.component import Component
from dagster.components.core.defs_module import ComponentRequirementsModel
from dagster.components.core.package_entry import (
    ComponentsEntryPointLoadError,
    discover_entry_point_package_objects,
    discover_package_objects,
    get_plugin_entry_points,
)
from dagster.components.core.snapshot import get_package_entry_snap


@click.group(name="list")
def list_cli():
    """Commands for listing Dagster components and related entities."""


@list_cli.command(name="plugins")
@click.option("--entry-points/--no-entry-points", is_flag=True, default=True)
@click.argument("extra_modules", nargs=-1, type=str)
def list_plugins_command(entry_points: bool, extra_modules: tuple[str]) -> None:
    """List registered plugin objects."""
    click.echo(serialize_value(list_plugins(entry_points, list(extra_modules))))


def list_plugins(
    entry_points: bool, extra_modules: Sequence[str]
) -> Union[PluginManifest, SerializableErrorInfo]:
    modules = [*(ep.value for ep in get_plugin_entry_points()), *extra_modules]
    try:
        plugin_objects = _load_plugin_objects(entry_points, extra_modules)
        object_snaps = [get_package_entry_snap(key, obj) for key, obj in plugin_objects.items()]
        return PluginManifest(
            modules=modules,
            objects=object_snaps,
        )
    except ComponentsEntryPointLoadError as e:
        tb = TracebackException.from_exception(e)
        return SerializableErrorInfo.from_traceback(tb)


@list_cli.command(name="all-components-schema")
@click.option("--entry-points/--no-entry-points", is_flag=True, default=True)
@click.argument("extra_modules", nargs=-1, type=str)
def list_all_components_schema_command(entry_points: bool, extra_modules: tuple[str, ...]) -> None:
    """Builds a JSON schema which ORs the schema for a component
    file for all component types available in the current code location.
    """
    click.echo(json.dumps(list_all_components_schema(entry_points, extra_modules)))


def list_all_components_schema(
    entry_points: bool, extra_modules: tuple[str, ...]
) -> dict[str, Any]:
    component_types = _load_component_types(entry_points, extra_modules)
    model_cls_list = []
    for key in sorted(component_types.keys(), key=lambda k: k.to_typename()):
        component_type = component_types[key]
        # Create ComponentFileModel schema for each type
        model_cls = component_type.get_model_cls()
        key_string = key.to_typename()
        if model_cls:
            model_cls_list.append(
                create_model(
                    key.name,
                    type=(Literal[key_string], key_string),
                    attributes=(model_cls, None),
                    requirements=(Optional[ComponentRequirementsModel], None),
                    __config__=ConfigDict(extra="forbid"),
                )
            )
    union_type = Union[tuple(model_cls_list)]  # type: ignore
    return TypeAdapter(union_type).json_schema()


@list_cli.command(name="definitions")
@python_pointer_options
@click.option(
    "--location", "-l", help="Name of the code location, can be used to scope environment variables"
)
@click.option(
    "--output-file",
    help="Write to file instead of stdout. If not specified, will write to stdout.",
)
@click.pass_context
def list_definitions_command(
    ctx: click.Context,
    location: Optional[str],
    output_file: Optional[str],
    **other_opts: object,
) -> None:
    """List Dagster definitions."""
    all_defs = list_definitions_impl(location, **other_opts)
    output = serialize_value(all_defs)
    if output_file:
        click.echo("[dagster-components] Writing to file " + output_file)
        Path(output_file).write_text(output)
    else:
        click.echo(output)


def list_definitions_impl(
    location: Optional[str],
    **other_opts: object,
) -> list[DgDefinitionMetadata]:
    python_pointer_opts = PythonPointerOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)

    with get_possibly_temporary_instance_for_cli(
        "``dagster-components definitions list``"
    ) as instance:
        instance.inject_env_vars(location)

        logger = logging.getLogger("dagster")

        removed_system_frame_hint = (
            lambda is_first_hidden_frame,
            i: f"  [{i} dagster system frames hidden, run dg check defs --verbose to see the full stack trace]\n"
            if is_first_hidden_frame
            else f"  [{i} dagster system frames hidden]\n"
        )

        try:
            repository_origin = get_repository_python_origin_from_cli_opts(python_pointer_opts)
            recon_repo = recon_repository_from_origin(repository_origin)
            repo_def = recon_repo.get_definition()
        except Exception:
            underlying_error = remove_system_frames_from_error(
                serializable_error_info_from_exc_info(sys.exc_info()),
                build_system_frame_removed_hint=removed_system_frame_hint,
            )

            logger.error(f"Loading location {location} failed:\n\n{underlying_error.to_string()}")
            sys.exit(1)

        all_defs: list[DgDefinitionMetadata] = []

        asset_graph = repo_def.asset_graph
        for key in sorted(
            list(asset_graph.get_all_asset_keys()), key=lambda key: key.to_user_string()
        ):
            node = asset_graph.get(key)
            all_defs.append(
                DgAssetMetadata(
                    key=key.to_user_string(),
                    deps=sorted([k.to_user_string() for k in node.parent_keys]),
                    group=node.group_name,
                    kinds=sorted(list(node.kinds)),
                    description=node.description,
                    automation_condition=node.automation_condition.get_label()
                    if node.automation_condition
                    else None,
                )
            )
        for key in asset_graph.asset_check_keys:
            node = asset_graph.get(key)
            all_defs.append(
                DgAssetCheckMetadata(
                    key=key.to_user_string(),
                    asset_key=key.asset_key.to_user_string(),
                    name=key.name,
                    additional_deps=sorted([k.to_user_string() for k in node.parent_entity_keys]),
                    description=node.description,
                )
            )
        for job in repo_def.get_all_jobs():
            if not is_reserved_asset_job_name(job.name):
                all_defs.append(DgJobMetadata(name=job.name))
        for schedule in repo_def.schedule_defs:
            schedule_str = (
                schedule.cron_schedule
                if isinstance(schedule.cron_schedule, str)
                else ", ".join(schedule.cron_schedule)
            )
            all_defs.append(
                DgScheduleMetadata(
                    name=schedule.name,
                    cron_schedule=schedule_str,
                )
            )
        for sensor in repo_def.sensor_defs:
            all_defs.append(DgSensorMetadata(name=sensor.name))
        for name, resource in repo_def.get_top_level_resources().items():
            all_defs.append(DgResourceMetadata(name=name, type=get_resource_type_name(resource)))

        return all_defs


# ########################
# ##### HELPERS
# ########################


def _load_plugin_objects(
    entry_points: bool, extra_modules: Sequence[str]
) -> dict[PluginObjectKey, object]:
    objects = {}
    if entry_points:
        objects.update(discover_entry_point_package_objects())
    if extra_modules:
        objects.update(discover_package_objects(extra_modules))
    return objects


def _load_component_types(
    entry_points: bool, extra_modules: Sequence[str]
) -> dict[PluginObjectKey, type[Component]]:
    return {
        key: obj
        for key, obj in _load_plugin_objects(entry_points, extra_modules).items()
        if isinstance(obj, type) and issubclass(obj, Component)
    }



from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from typing import Any, Optional, TypeVar

from dagster_shared.serdes import serialize_value

from dagster import _check as check
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.definitions_load_context import (
    DefinitionsLoadContext,
    DefinitionsLoadType,
)
from dagster._core.definitions.metadata.metadata_value import (
    CodeLocationReconstructionMetadataValue,
)
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
    RepositoryLoadData,
)
from dagster.components.definitions import (
    LazyDefinitions as LazyDefinitions,
    definitions as definitions,
)

T_Defs = TypeVar("T_Defs", Definitions, RepositoryDefinition)


@contextmanager
def scoped_definitions_load_context(
    load_type: DefinitionsLoadType = DefinitionsLoadType.INITIALIZATION,
    repository_load_data: Optional[RepositoryLoadData] = None,
) -> Iterator["DefinitionsLoadContext"]:
    context = DefinitionsLoadContext(load_type=load_type, repository_load_data=repository_load_data)
    curr_context = DefinitionsLoadContext.get()
    try:
        DefinitionsLoadContext.set(context)
        yield context
    finally:
        DefinitionsLoadContext.set(curr_context)


@contextmanager
def scoped_reconstruction_serdes_objects(
    state_objects: Optional[Mapping[str, Any]] = None,
) -> Iterator[None]:
    """For test cases that involved state-backed definition objects. This context manager
    allows one to set backing state for a definitions object and test reconstruction
    logic. Creates a DefinitionsLoadContext with DefinitionsLoadType.RECONSTRUCTION,
    serializes the passed in objects and wraps them in
    a CodeLocationReconstructionMetadataValue on your behalf.

    Examples:

    .. code-block:: python

        with scoped_reconstruction_serdes_objects(test_key=ExampleDefState(a_string="bar")):
            loader_cached = ExampleStateBackedDefinitionsLoader("test_key")
            defs = loader_cached.build_defs()
            assert len(defs.get_all_asset_specs()) == 1
            assert defs.get_assets_def("bar")
    """
    with scoped_reconstruction_metadata(
        {k: serialize_value(v) for k, v in state_objects.items()} if state_objects else None
    ):
        yield


@contextmanager
def scoped_reconstruction_metadata(
    reconstruction_metadata: Optional[Mapping[str, Any]] = None,
) -> Iterator[None]:
    """For test cases that involved state-backed definition objects. This context manager
    allows one to set backing reconstruciton metaddata for a definitions object
    and test reconstruction logic. Creates a DefinitionsLoadContext with
    DefinitionsLoadType.RECONSTRUCTION. Wraps each passed value in a
    CodeLocationReconstructionMetadataValue.

    Examples:

    .. code-block:: python

        with scoped_reconstruction_metadata({"test_key": "test_value"}):
            loader_cached = ExampleStateBackedDefinitionsLoader("test_key")
            defs = loader_cached.build_defs()
            assert len(defs.get_all_asset_specs()) == 1
            assert defs.get_assets_def("bar")
    """
    prev_context = DefinitionsLoadContext.get()
    reconstruction_metadata = reconstruction_metadata or {}
    try:
        prev_load_data = prev_context._repository_load_data  # noqa
        DefinitionsLoadContext.set(
            DefinitionsLoadContext(
                load_type=DefinitionsLoadType.RECONSTRUCTION,
                repository_load_data=RepositoryLoadData(
                    cacheable_asset_data=prev_load_data.cacheable_asset_data
                    if prev_load_data
                    else {},
                    reconstruction_metadata={
                        **{
                            k: CodeLocationReconstructionMetadataValue(v)
                            for k, v in reconstruction_metadata.items()
                        },
                        **(prev_load_data.reconstruction_metadata if prev_load_data else {}),
                    },
                ),
            )
        )
        yield
    finally:
        DefinitionsLoadContext.set(prev_context)


def unwrap_reconstruction_metadata(repo_def: RepositoryDefinition) -> Mapping[str, Any]:
    """Takes the metadata from a Definitions object and unwraps the CodeLocationReconstructionMetadataValue
    metadata values into a dictionary.
    """
    return {
        k: metadata_value.value
        for k, metadata_value in check.not_none(
            repo_def.repository_load_data
        ).reconstruction_metadata.items()
        if isinstance(metadata_value, CodeLocationReconstructionMetadataValue)
    }



from abc import abstractmethod

from dagster.components.scaffold.scaffold import Scaffolder, ScaffoldRequest, TModel


class ShimScaffolder(Scaffolder[TModel]):
    @abstractmethod
    def get_text(self, request: ScaffoldRequest[TModel]) -> str: ...

    def scaffold(self, request: ScaffoldRequest[TModel]) -> None:
        if request.target_path.suffix != ".py":
            raise ValueError("Invalid target path suffix. Expected a path ending in `.py`.")
        # temporary hack as currently all scaffold requests target directories
        # that are auto-created
        request.target_path.rmdir()
        request.target_path.write_text(self.get_text(request))



from dagster._core.definitions.decorators.schedule_decorator import schedule
from dagster.components.lib.shim_components.base import ShimScaffolder
from dagster.components.scaffold.scaffold import ScaffoldRequest, scaffold_with


class ScheduleScaffolder(ShimScaffolder):
    def get_text(self, request: ScaffoldRequest) -> str:
        return f"""from typing import Union

import dagster as dg


@dg.schedule(cron_schedule="@daily", target="*")
def {request.target_path.stem}(context: dg.ScheduleEvaluationContext) -> Union[dg.RunRequest, dg.SkipReason]:
    return dg.SkipReason("Skipping. Change this to return a RunRequest to launch a run.")
"""


scaffold_with(ScheduleScaffolder)(schedule)



import inspect
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from dagster_shared.record import IHaveNew, record_custom
from dagster_shared.yaml_utils.source_position import SourcePosition
from pydantic import BaseModel, TypeAdapter
from typing_extensions import Self

import dagster._check as check
from dagster._annotations import PublicAttr, preview, public
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.metadata.source_code import CodeReference, LocalFileCodeReference
from dagster._core.definitions.utils import validate_component_owner
from dagster.components.component.component_scaffolder import DefaultComponentScaffolder
from dagster.components.resolved.base import Resolvable
from dagster.components.scaffold.scaffold import scaffold_with

if TYPE_CHECKING:
    from dagster.components.core.context import ComponentLoadContext


@public
@preview(emit_runtime_warning=False)
@record_custom
class ComponentTypeSpec(IHaveNew):
    """Specifies the core attributes of a component.

    Args:
        description (Optional[str]): Human-readable description of this component.
        metadata (Optional[Dict[str, Any]]): A dict of static metadata for this component.
            For example, users can provide information about the database table this
            component corresponds to.
        owners (Optional[Sequence[str]]): A list of strings representing owners of the component. Each
            string can be a user's email address, or a team name prefixed with `team:`,
            e.g. `team:finops`.
        tags (Optional[Sequence[str]]): Tags for filtering and organizing.

    """

    description: PublicAttr[Optional[str]]
    tags: PublicAttr[Sequence[str]]
    owners: PublicAttr[Sequence[str]]
    metadata: PublicAttr[Mapping[str, Any]]

    def __new__(
        cls,
        description: Optional[str] = None,
        tags: Optional[Sequence[str]] = None,
        owners: Optional[Sequence[str]] = None,
        metadata: Optional[Mapping[str, Any]] = None,
    ):
        owners = check.opt_sequence_param(owners, "owners", of_type=str)
        for owner in owners:
            validate_component_owner(owner)

        return super().__new__(
            cls,
            description=check.opt_str_param(description, "description"),
            tags=check.opt_sequence_param(tags, "tags", of_type=str),
            owners=owners,
            metadata=check.opt_mapping_param(metadata, "metadata", key_type=str),
        )


@public
@preview(emit_runtime_warning=False)
@scaffold_with(DefaultComponentScaffolder)
class Component(ABC):
    """Components are a tool for dynamically creating Dagster definitions.
    A Component subclass must implement the build_defs method. It may also
    inherit from Resolvable or implement get_model_cls to support instantiation
    via yaml.
    """

    @classmethod
    def __dg_package_entry__(cls) -> None: ...

    @classmethod
    def get_schema(cls) -> Optional[type[BaseModel]]:
        return None

    @classmethod
    def get_spec(cls) -> ComponentTypeSpec:
        return ComponentTypeSpec()

    @classmethod
    def get_model_cls(cls) -> Optional[type[BaseModel]]:
        if issubclass(cls, Resolvable):
            return cls.model()

        # handle existing overrides for backwards compatibility
        cls_from_get_schema = cls.get_schema()
        if cls_from_get_schema:
            return cls_from_get_schema

        return None

    @classmethod
    def get_additional_scope(cls) -> Mapping[str, Any]:
        return {}

    @abstractmethod
    def build_defs(self, context: "ComponentLoadContext") -> Definitions: ...

    @classmethod
    def load(cls, attributes: Optional[BaseModel], context: "ComponentLoadContext") -> Self:
        if issubclass(cls, Resolvable):
            return (
                cls.resolve_from_model(
                    context.resolution_context.at_path("attributes"),
                    attributes,
                )
                if attributes
                else cls()
            )
        else:
            # If the Component does not implement anything from Resolved, try to instantiate it without
            # argument.
            return cls()

    @classmethod
    def get_code_references_for_yaml(
        cls, yaml_path: Path, source_position: SourcePosition, context: "ComponentLoadContext"
    ) -> Sequence[CodeReference]:
        """Returns a list of code references for a component which has been defined in a YAML file.

        Args:
            yaml_path (Path): The path to the YAML file where this component is defined.
            source_position (SourcePosition): The source position of the component in the YAML file.
            context (ComponentLoadContext): The context in which the component is being loaded.
        """
        return [
            LocalFileCodeReference(file_path=str(yaml_path), line_number=source_position.start.line)
        ]

    @classmethod
    def get_description(cls) -> Optional[str]:
        return cls.get_spec().description or inspect.getdoc(cls)

    @classmethod
    def from_attributes_dict(
        cls, *, attributes: dict, context: Optional["ComponentLoadContext"] = None
    ) -> Self:
        """Load a Component from a dictionary. The dictionary is what would exist in the component.yaml file
        under the "attributes" key.

        Examples:

        .. code-block:: python

            class ModelComponentWithDeclaration(Component, Model, Resolvable):
                value: str

                def build_defs(self, context: ComponentLoadContext) -> Definitions: ...

            assert (
                component_defs(
                    component=ModelComponentWithDeclaration.from_attributes_dict(attributes={"value": "foobar"}),
                ).get_assets_def("an_asset")()
                == "foobar"
            )

        Args:
            attributes (dict): The attributes to load the Component from.
            context (Optional[ComponentLoadContext]): The context to load the Component from.

        Returns:
            A Component instance.
        """
        from dagster.components.core.context import ComponentLoadContext

        model_cls = cls.get_model_cls()
        assert model_cls
        model = TypeAdapter(model_cls).validate_python(attributes)
        return cls.load(model, context if context else ComponentLoadContext.for_test())

    @classmethod
    def from_yaml_path(
        cls, yaml_path: Path, context: Optional["ComponentLoadContext"] = None
    ) -> "Component":
        """Load a Component from a yaml file.

        Args:
            yaml_path (Path): The path to the yaml file.
            context (Optional[ComponentLoadContext]): The context to load the Component from. Defaults to a test context.

        Returns:
            A Component instance.
        """
        from dagster.components.core.context import ComponentLoadContext
        from dagster.components.core.defs_module import load_yaml_component_from_path

        return load_yaml_component_from_path(
            context=context or ComponentLoadContext.for_test(), component_def_path=yaml_path
        )
